#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for schooltool.relationship.objectevents
"""
import unittest
import doctest

from schooltool.relationship.tests import URIStub


def doctest_delete_breaks_relationships():
    """When you delete an object, all of its relationships should be removed

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()

        >>> import zope.event
        >>> old_subscribers = zope.event.subscribers[:]
        >>> from schooltool.relationship.objectevents import unrelateOnDeletion
        >>> zope.event.subscribers.append(unrelateOnDeletion)

    Suppose we have two related objects

        >>> from schooltool.relationship.tests import SomeObject
        >>> apple = SomeObject('apple')
        >>> orange = SomeObject('orange')

        >>> from schooltool.relationship import getRelatedObjects, relate
        >>> relate('example:Relationship',
        ...             (apple, URIStub('example:One')),
        ...             (orange, URIStub('example:Two')))
        >>> getRelatedObjects(apple, URIStub('example:Two'))
        [orange]

    We put those objects to a Zope 3 container.

        >>> from zope.container.btree import BTreeContainer
        >>> container = BTreeContainer()
        >>> container['apple'] = apple
        >>> container['orange'] = orange

    When we delete an object, all of its relationships should disappear

        >>> del container['orange']
        >>> getRelatedObjects(apple, URIStub('example:Two'))
        []

        >>> zope.event.subscribers[:] = old_subscribers
        >>> tearDown()

    """


def doctest_copy_breaks_relationships():
    """When you copy an object, all of its relationships should be removed

    (An alternative solution would be to clone the relationships, but I'm
    wary of that path.  What happens if you copy and paste objects between
    different application instances?)

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()

        >>> import zope.event
        >>> old_subscribers = zope.event.subscribers[:]
        >>> from schooltool.relationship.objectevents import unrelateOnCopy
        >>> zope.event.subscribers.append(unrelateOnCopy)

    Suppose we have two related objects.  We must have objects that are
    IContained, otherwise ObjectCopier will happily duplicate all related
    objects as well as relationship links.

        >>> from schooltool.relationship.tests import SomeContained
        >>> apple = SomeContained('apple')
        >>> orange = SomeContained('orange')

        >>> from schooltool.relationship import getRelatedObjects, relate
        >>> relate('example:Relationship',
        ...             (apple, URIStub('example:One')),
        ...             (orange, URIStub('example:Two')))
        >>> getRelatedObjects(apple, URIStub('example:Two'))
        [orange]

    We put those objects to a Zope 3 container.

        >>> from zope.container.btree import BTreeContainer
        >>> container = BTreeContainer()
        >>> container['apple'] = apple
        >>> container['orange'] = orange

    We copy one of the objects to another container.

        >>> from zope.copypastemove import ObjectCopier
        >>> another_container = BTreeContainer()
        >>> copier = ObjectCopier(container['orange'])
        >>> new_name = copier.copyTo(another_container)
        >>> copy_of_orange = another_container[new_name]

    When we copy an object, all of its relationships should disappear

        >>> from schooltool.relationship.interfaces import IRelationshipLinks
        >>> list(IRelationshipLinks(copy_of_orange))
        []

    The old relationships should still work

        >>> getRelatedObjects(apple, URIStub('example:Two'))
        [orange]
        >>> getRelatedObjects(orange, URIStub('example:One'))
        [apple]

        >>> zope.event.subscribers[:] = old_subscribers
        >>> tearDown()

    """


def doctest_copy_does_not_break_inside_relationships():
    """When you copy an object, all of its relationships should be removed

    This is a regression test for the following bug: If x is related to x.y
    where x.y is a subobject of x, when you copy x to x', all links from x'
    will be removed, but the copied link on x'.y' will remain.

        >>> from schooltool.relationship.tests import setUp, tearDown
        >>> setUp()

        >>> import zope.event
        >>> old_subscribers = zope.event.subscribers[:]
        >>> from schooltool.relationship.objectevents import unrelateOnCopy
        >>> zope.event.subscribers.append(unrelateOnCopy)

    Suppose we have two related objects.  We must have objects that are
    IContained, otherwise ObjectCopier will happily duplicate all related
    objects as well as relationship links.

        >>> from schooltool.relationship.tests import SomeContained
        >>> apple = SomeContained('apple')
        >>> orange = SomeContained('orange')

        >>> from schooltool.relationship import getRelatedObjects, relate
        >>> relate(URIStub('example:Relationship'),
        ...             (apple, URIStub('example:One')),
        ...             (orange, URIStub('example:Two')))
        >>> getRelatedObjects(apple, URIStub('example:Two'))
        [orange]

    We put one of the objects to a Zope 3 container.

        >>> from zope.container.btree import BTreeContainer
        >>> container = BTreeContainer()
        >>> container['orange'] = orange

    We make the other object a subobject of the first object

        >>> apple.__parent__ = orange
        >>> apple.__name__ = 'apple'
        >>> orange.apple = apple

    We copy the first object to another container.

        >>> from zope.copypastemove import ObjectCopier
        >>> another_container = BTreeContainer()
        >>> copier = ObjectCopier(container['orange'])
        >>> new_name = copier.copyTo(another_container)
        >>> copy_of_orange = another_container[new_name]
        >>> copy_of_apple = copy_of_orange.apple

        >>> r1 = URIStub('example:One')
        >>> r2 = URIStub('example:Two')

    When we copy an object, its internal relationships should remain

        >>> getRelatedObjects(copy_of_orange, r1)
        [apple]
        >>> getRelatedObjects(copy_of_apple, r2)
        [orange]

    These two objects are, in fact, copies

        >>> getRelatedObjects(copy_of_orange, r1)[0] is copy_of_apple
        True
        >>> getRelatedObjects(copy_of_apple, r2)[0] is copy_of_orange
        True

    The old relationships should still work

        >>> getRelatedObjects(apple, r2)
        [orange]
        >>> getRelatedObjects(orange, r1)
        [apple]

        >>> zope.event.subscribers[:] = old_subscribers
        >>> tearDown()

    """


def test_suite():
    return unittest.TestSuite([
                doctest.DocTestSuite(),
           ])

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool PDF report views.
"""

import urllib
from datetime import datetime

from zope.interface import implements
from zope.cachedescriptors.property import Lazy
from zope.component import getMultiAdapter
from reportlab.lib import units
from reportlab.lib import pagesizes
from zope.publisher.browser import BrowserView
from zope.i18n import translate
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from z3c.rml import rml2pdf

from schooltool.common import SchoolToolMessage as _
from schooltool.common.inlinept import InlinePageTemplate
from schooltool.table.column import getResourceURL
from schooltool.app import pdf
from schooltool.app.browser.interfaces import IReportPageTemplate


def _quoteUrl(url):
    if not url:
        return url
    if type(url) is unicode:
        encoded = url.encode('UTF-8')
    else:
        encoded = str(url)
    return urllib.quote(encoded)


class PDFView(BrowserView):

    template = None # Page template for rendering RML
    filename = '' # Suggested PDF file name
    inline = False # Display PDF in browser? (otherwise download)

    pdf_disabled_text = _("PDF support is disabled."
                          "  It can be enabled by your administrator.")

    def setUpResponse(self, data, filename):
        response = self.request.response
        response.setHeader('Content-Type', 'application/pdf')
        response.setHeader('Content-Length', len(data))
        # We don't really accept ranges, but Acrobat Reader will not show the
        # report in the browser page if this header is not provided.
        response.setHeader('Accept-Ranges', 'bytes')

        disposition = self.inline and 'inline' or 'attachment'
        if filename:
            disposition += '; filename="%s"' % filename
        response.setHeader('Content-Disposition', disposition)

    def renderToFile(self):
        filename = _quoteUrl(self.filename)
        if not pdf.isEnabled():
            return filename, None
        filename = _quoteUrl(self.filename)
        xml = self.template()
        stream = rml2pdf.parseString(xml, filename=filename)
        data = stream.getvalue()
        return filename, data

    def __call__(self):
        if not pdf.isEnabled():
            return translate(self.pdf_disabled_text, context=self.request)
        filename, data = self.renderToFile()
        self.setUpResponse(data, filename)
        return data

class PageTemplateEnablingHelper(object):

    def __init__(self, view):
        self.view = view

    def __getitem__(self, template_name):
        templates = dict(self.view.page_templates)
        if template_name in templates:
            return ''
        page_template = getMultiAdapter(
            (self.view.context, self.view.request, self.view),
            IReportPageTemplate,
            name=template_name)
        self.view.page_templates.append((template_name, page_template))
        return page_template


class ReportPageTemplate(object):
    implements(IReportPageTemplate)

    template = None # Page template for rendering RML page template tag
    style_template = None # Page template for rendering RML stylesheet contents

    def __init__(self, context, request, view):
        self.context, self.request = context, request
        self.parent = view

    def stylesheet(self):
        if self.style_template is not None:
            return self.style_template(
                view=self, context=self.context, request=self.request)
        return ''

    def __call__(self):
        return self.template()


class DefaultPageTemplate(ReportPageTemplate):

    template=ViewPageTemplateFile('templates/default_report_template.pt',
                                  content_type="text/xml")
    style_template = InlinePageTemplate("""
      <paraStyle name="_footer_page_number"
        xmlns:tal="http://xml.zope.org/namespaces/tal"
        alignment="center" fontName="Times_New_Roman"
        tal:attributes="fontSize view/footer/font_height" />
    """, content_type="text/xml")

    footer_text = u"" # additional short footer text
    report_date = None # report generation date

    def __init__(self, *args, **kw):
        ReportPageTemplate.__init__(self, *args, **kw)
        self.report_date = datetime.today()

    @property
    def header(self):
        if not self.parent.title:
            return {'height': 0}
        doc_w, doc_h = self.parent.pageSize
        font_height = 0.25 * units.inch
        header_height = font_height + 0.2 * units.inch
        return {
            'title': {
                'x': (doc_w/2.0),
                'y': (doc_h - self.parent.topMargin - font_height),
                },
            'height': header_height,
            }

    @property
    def footer(self):
        doc_w, doc_h = self.parent.pageSize

        footer_height = 0.6 * units.inch
        font_height = 10 # ~0.14 inch
        rule_y = (self.parent.bottomMargin + 0.55 * units.inch)

        return {
            'rule': '%d %d %d %d' % (
                self.parent.leftMargin, rule_y,
                doc_w - self.parent.rightMargin, rule_y),
            'logo': {
                'x': self.parent.leftMargin,
                'y': self.parent.bottomMargin,
                'height': 0.5 * units.inch,
                },
            'font_height': font_height,
            'center_place': {
                'x': self.parent.leftMargin,
                'y': (self.parent.bottomMargin + footer_height
                      - 0.2 * units.inch
                      - font_height),
                'width': (doc_w - self.parent.leftMargin - self.parent.rightMargin),
                'height': font_height + 0.1 * units.inch,
                },
            'right_str': {
                'x': (doc_w - self.parent.rightMargin),
                'y': (self.parent.bottomMargin + footer_height
                      - 0.1 * units.inch - font_height),
                },
            'right_str_2': {
                'x': (doc_w - self.parent.rightMargin),
                'y': (self.parent.bottomMargin + footer_height
                      - 0.15 * units.inch - font_height * 2),
                },
            'height': footer_height,
            }

    @property
    def frame(self):
        doc_w, doc_h = self.parent.pageSize
        return {
            'x': self.parent.leftMargin,
            'y': self.parent.bottomMargin + self.footer['height'],
            'width': doc_w - self.parent.leftMargin - self.parent.rightMargin,
            'height': (doc_h - self.parent.topMargin - self.parent.bottomMargin
                       - self.header['height'] - self.footer['height']),
            }


class Box(object):
    """XXX: Scheduled for demolition"""

    def __init__(self, top, right=None, bottom=None, left=None):
        self.top = top
        if right is None:
            assert bottom is None, "Need to set right"
            assert left is None, "Need to set right"
            right = top
        self.right = right
        if left is None:
            self.left = right
            if bottom is None:
                self.bottom = top
            else:
                self.bottom = bottom
        else:
            assert bottom is not None, "Need to set bottom"
            self.left = left
            self.bottom = bottom


class ReportPDFView(PDFView):
    leftMargin = 0.25 * units.inch
    rightMargin = 0.25 * units.inch
    topMargin = 0.25 * units.inch
    bottomMargin = 0.25 * units.inch
    margin = Box(topMargin, rightMargin, bottomMargin, leftMargin)
    rotation = 0
    pageSize = pagesizes.A4
    title = u""

    def __init__(self, *args, **kw):
        PDFView.__init__(self, *args, **kw)
        self.page_templates = []
        self.use_template = PageTemplateEnablingHelper(self)


class FlourishDefaultPageTemplate(ReportPageTemplate):
    """XXX: Scheduled for demolition"""

    template = ViewPageTemplateFile(
        'templates/f_default_report_template.pt',
        content_type='text/xml')
    style_template = ViewPageTemplateFile(
        'templates/f_default_style_template.pt',
        content_type='text/xml')

    def lines(self, attr, top, left):
        content = attr['content']
        if not isinstance(content, list):
            content = [content]
        content = filter(None, content)
        margin = attr['margin']
        result = []
        line_height = attr['fontSize'] + margin.top + margin.bottom
        for i, line in enumerate(content):
            result.append({
                    'content': line,
                    'x': left,
                    'y': top - (line_height * (i+1)) + margin.bottom,
                    })
        return line_height * len(result), result

    @Lazy
    def top_line(self):
        width = self.frame['width']
        x1 = self.frame['x']
        x2 = x1 + width
        y = self.header['y'] - (self.frame['margin'].top / 2.0)
        return {
            'coords': '%d %d %d %d' % (x1, y, x2, y),
            'color': '#000000',
            'style': 'square',
            }

    @Lazy
    def bottom_line(self):
        width = self.frame['width']
        x1 = self.frame['x']
        x2 = x1 + width
        y = self.frame['y'] - (self.frame['margin'].top / 2.0)
        return {
            'coords': '%d %d %d %d' % (x1, y, x2, y),
            'color': '#000000',
            'style': 'square',
            }

    @Lazy
    def header(self):
        doc_w, doc_h = self.parent.pageSize
        title = {
            'fontSize': 24,
            'margin': Box(0, 0, 8, 0),
            'content': self.parent.title,
            }
        subtitle = {
            'fontSize': 12,
            'margin': Box(0, 0, 6, 0),
            'content': getattr(self.parent, 'subtitle'),
            }
        top = self.top_bar['y']
        left = self.parent.margin.left
        title_height, title_lines = self.lines(title, top, left)
        title['lines'] = title_lines
        top -= title_height
        subtitle_height, subtitle_lines = self.lines(subtitle, top, left)
        subtitle['lines'] = subtitle_lines
        height = title_height + subtitle_height
        width = doc_w - self.parent.margin.left - self.parent.margin.right
        x = self.parent.margin.left
        y = self.top_bar['y'] - height
        backgroundColor = '#ffffff'
        color = '#000000'
        return {
            'backgroundColor': backgroundColor,
            'color': color,
            'title': title,
            'subtitle': subtitle,
            'height': height,
            'width': width,
            'x': x,
            'y': y,
            }

    @Lazy
    def top_bar(self):
        fontSize = 12
        padding = Box(8, 10.5)
        height = fontSize + padding.top + padding.bottom
        doc_w, doc_h = self.parent.pageSize
        width = doc_w - self.parent.margin.left - self.parent.margin.right
        x = self.parent.margin.left
        y = doc_h - self.parent.margin.top - height
        slot_y = y + padding.bottom
        backgroundColor = '#636466'
        color = '#ffffff'
        return {
            'backgroundColor': backgroundColor,
            'color': color,
            'fontSize': fontSize,
            'height': height,
            'width': width,
            'x': x,
            'y': y,
            'slots': {
                'left': {
                    'x': x + padding.left,
                    'y': slot_y,
                    },
                'center': {
                    'x': x + padding.left,
                    'y': slot_y,
                    },
                'right': {
                    'x': x + width - padding.right,
                    'y': slot_y,
                    },
                }
            }

    @Lazy
    def bottom_bar(self):
        fontSize = 8.5
        padding = Box(1.5, 10.5)
        height = fontSize + padding.top + padding.bottom
        doc_w, doc_h = self.parent.pageSize
        width = doc_w - self.parent.margin.left - self.parent.margin.right
        x = self.parent.margin.left
        y = self.parent.margin.bottom
        slot_y = y + padding.bottom
        backgroundColor = '#ffffff'
        color = '#000000'
        url = getResourceURL('schooltool.skin.flourish',
                             'logo_bw.png',
                             self.request)
        return {
            'logo_url': url,
            'backgroundColor': backgroundColor,
            'color': color,
            'fontSize': fontSize,
            'height': height,
            'width': width,
            'x': x,
            'y': y,
            'slots': {
                'left': {
                    'x': x + padding.left,
                    'y': slot_y,
                    },
                'center': {
                    'x': x + padding.left,
                    'y': slot_y,
                    },
                'right': {
                    'x': x + width - padding.right,
                    'y': slot_y,
                    },
                }
            }

    @Lazy
    def frame(self):
        doc_w, doc_h = self.parent.pageSize
        margin = Box(8, 0)
        width = (doc_w - self.parent.margin.left - self.parent.margin.right
                 - margin.left - margin.right)
        height = (doc_h - self.parent.margin.top - self.parent.margin.bottom
                  - self.top_bar['height'] - self.bottom_bar['height']
                  - self.header['height']
                  - margin.top - margin.bottom)
        x = self.parent.margin.left + margin.left
        y = (self.parent.margin.bottom + self.bottom_bar['height'] +
             margin.bottom)
        return {
            'height': height,
            'margin': margin,
            'width': width,
            'x': x,
            'y': y,
            }


class FlourishReportPDFView(PDFView):
    """XXX: Scheduled for demolition"""

    margin = Box(0.75*units.inch)
    rotation = 0
    pageSize = pagesizes.A4
    title = u''
    subtitle = u''
    slots = {
        'top_left': u'',
        'top_center': u'',
        'top_right': u'',
        }

    def __init__(self, *args, **kw):
        super(FlourishReportPDFView, self).__init__(*args, **kw)
        self.page_templates = []
        self.use_template = PageTemplateEnablingHelper(self)



#
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Schooltool grade level views.
"""

from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.container.interfaces import INameChooser
from zope.component import adapts
from zope.interface import implements
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.browser import BrowserView
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.traversing.browser.absoluteurl import absoluteURL

from z3c.form import form, field, button

import schooltool.skin.flourish.page
import schooltool.skin.flourish.form
import schooltool.skin.flourish.content
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.table.table import simple_form_key
from schooltool.level.interfaces import ILevel, ILevelContainer
from schooltool.level.level import Level
from schooltool.skin import flourish

from schooltool.common import SchoolToolMessage as _


class LevelAddForm(form.AddForm):
    """Contact add form for basic contact."""

    label = _("Add new level")
    template = ViewPageTemplateFile('templates/add_level_subform.pt')
    fields = field.Fields(ILevel)

    def updateActions(self):
        super(LevelAddForm, self).updateActions()
        self.actions['add'].addClass('button-ok')

    @button.buttonAndHandler(_("Add"), name='add')
    def handleAdd(self, action):
        # Pretty, isn't it?
        form.AddForm.handleAdd.func(self, action)

    def create(self, data):
        level = Level()
        form.applyChanges(self, level, data)
        return level

    def add(self, level):
        container = self.getContent()
        name = INameChooser(container).chooseName('', level)
        container[name] = level

    @button.buttonAndHandler(_("Cancel"), name='cancel')
    def handle_cancel_action(self, action):
        self.request.response.redirect(self.nextURL())

    def nextURL(self):
        return absoluteURL(self.context, self.request)


class LevelEditView(form.EditForm):
    template = ViewPageTemplateFile('templates/level_edit.pt')
    fields = field.Fields(ILevel)

    def redirectToContainer(self):
        url = absoluteURL(self.context.__parent__, self.request)
        self.request.response.redirect(url)

    @button.buttonAndHandler(_("Apply"), name='apply')
    def handleApply(self, action):
        # Pretty, isn't it?
        form.EditForm.handleApply.func(self, action)
        self.redirectToContainer()

    @button.buttonAndHandler(_("Cancel"), name='cancel')
    def handle_cancel_action(self, action):
        self.redirectToContainer()

    def updateActions(self):
        super(LevelEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class LevelContainerView(BrowserView):
    template = ViewPageTemplateFile('templates/levels.pt')

    def __init__(self, context, request):
        super(LevelContainerView, self).__init__(context, request)
        self.addform = LevelAddForm(context, request)

    def redirectToContainer(self):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def levels(self):
        levels = self.context.values()
        result = []
        positions = range(len(levels))
        for n, level in enumerate(levels):
            pos_numbers = [
                {'value': pos+1,
                 'selected': (pos==n and 'selected' or None)}
                for pos in positions]
            result.append({
                'form_key': simple_form_key(level),
                'level': level,
                'positions': pos_numbers})
        return result

    def handleApply(self):
        keys = list(self.context.keys())
        new_keys = keys[:]
        for old_pos, level in enumerate(self.context.values()):
            form_key = 'pos.%s' % simple_form_key(level)
            if form_key not in self.request:
                continue
            new_pos = int(self.request[form_key]) - 1
            if new_pos != old_pos:
                new_keys.remove(level.__name__)
                new_keys.insert(new_pos, level.__name__)
        if keys != new_keys:
            self.context.updateOrder(new_keys)

    def handleDelete(self):
        formkey_to_name = dict(
            [(simple_form_key(level), level.__name__)
             for level in self.context.values()])
        for form_key in self.request.get('delete', []):
            if form_key in formkey_to_name:
                del self.context[formkey_to_name[form_key]]
        self.redirectToContainer()

    def update(self):
        if 'DELETE' in self.request:
            self.handleDelete()
        elif 'form-submitted' in self.request:
            self.handleApply()

    def __call__(self):
        self.update()
        return self.template()


class FlourishReorderLevelsView(flourish.page.Page, LevelContainerView):

    deleted = False

    def handleDelete(self):
        form_items = dict(
            [(simple_form_key(level), level.__name__)
             for level in self.context.values()])
        for form_key, name in form_items.items():
            if 'delete.%s' % form_key in self.request:
                del self.context[name]
                self.deleted = True

    def update(self):
        self.handleDelete()
        if self.deleted:
            self.request.response.redirect(self.request.URL)
        elif 'form-submitted' in self.request:
            self.handleApply()
            self.request.response.redirect(self.request.URL)


class LevelContainerAbsoluteURLAdapter(BrowserView):
    adapts(ILevelContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        app = ISchoolToolApplication(None)
        url = absoluteURL(app, self.request)
        return url + '/levels'

    __call__ = __str__


def appTitleContentFactory(context, request, view, name):
    app = ISchoolToolApplication(None)
    return flourish.content.queryContentProvider(
        app, request, view, 'title')


class FlourishLevelsView(flourish.page.Page):

    @property
    def title(self):
        app = ISchoolToolApplication(None)
        title = flourish.content.queryContentProvider(
            app, self.request, self, 'title')
        return title.title

    def table(self):
        result = []
        for level in list(self.context.values()):
            result.append({
               'level': level,
               })
        return result


class FlourishLevelAddView(flourish.form.AddForm, LevelAddForm):
    label = None
    legend = _('Level information')

    def updateActions(self):
        super(FlourishLevelAddView, self).updateActions()
        self.actions['cancel'].addClass('button-cancel')

class FlourishLevelEditView(flourish.form.Form, LevelEditView):
    label = None
    legend = _('Level information')


class LevelsAddLinks(flourish.page.RefineLinksViewlet):
    """Manager for Add links."""


class LevelsActionsLinks(flourish.page.RefineLinksViewlet):
    """Manager for Actions links."""



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for demographics fields.
"""
import unittest
import doctest

import zope.schema
from zope.interface.verify import verifyObject
from zope.publisher.browser import TestRequest
from zope.app.testing import setup
from zope.i18n import translate

import z3c.form

from schooltool.basicperson.browser.demographics import (
    CustomEnumDataConverter)


def doctest_CustomEnumDataConverter():
    r"""Tests for CustomEnumDataConverter.

    Value conversions between widget (multi-line unicode string) and
    field(list of text lines).

        >>> field = zope.schema.List(
        ...     __name__='values',
        ...     title=u'Values')

        >>> text = z3c.form.widget.Widget(TestRequest())

        >>> converter = CustomEnumDataConverter(field, text)

        >>> verifyObject(z3c.form.interfaces.IDataConverter, converter)
        True

    Empty field values produce an empty string for the widget.

        >>> converter.toWidgetValue(field.missing_value)
        u''

        >>> converter.toWidgetValue(None)
        u''

        >>> converter.toWidgetValue([])
        u''

    Field values must be iterable object of strings.

        >>> converter.toWidgetValue(('one', 2, '3'))
        Traceback (most recent call last):
        ...
        TypeError: sequence item 1: expected string, int found

        >>> converter.toWidgetValue(iter(['four', '5', 'six']))
        u'four\n5\nsix'

    Conversion to field knows how to filter out whitespace.

        >>> print converter.toFieldValue(u'')
        None

        >>> print converter.toFieldValue(u'  \n \t \t          \n')
        None

    Newlines separate values.

        >>> converter.toFieldValue('         single \t val  ')
        [u'single \t val']

        >>> converter.toFieldValue('one\ntwo\n3')
        [u'one', u'two', u'3']

        >>> converter.toFieldValue('''
        ...
        ...    We
        ...           even
        ...
        ...   support  crazy
        ...                    spacing ''')
        [u'We', u'even', u'support  crazy', u'spacing']

    Duplicate entries are not allowed.

        >>> from z3c.form.converter import FormatterValidationError
        >>> try:
        ...     converter.toFieldValue('''
        ...        department
        ...        of
        ...        redundancy
        ...        department
        ...     ''')
        ... except FormatterValidationError, e:
        ...     print translate(e.message)
        Duplicate entry "department"

    We also have a line length limit because we currently use IDNA encoding for
    vocabularies.

        >>> long_val = (u'Very, very long lines make the idna conversion '
        ...              'throw exceptions')

        >>> len(long_val.encode('idna'))
        63

        >>> converter.toFieldValue(long_val)
        [u'Very, very long lines make the idna conversion throw exceptions']

        >>> long_val += '!'

        >>> long_val.encode('idna')
        Traceback (most recent call last):
        ...
        UnicodeError: label empty or too long

        >>> try:
        ...     converter.toFieldValue(long_val)
        ... except FormatterValidationError, e:
        ...     print translate(e.message)
        Value too long "Very, very long lines make the
        idna conversion throw exceptions!"

    """


def setUp(test):
    setup.placelessSetUp()


def tearDown(test):
    setup.placelessTearDown()


def test_suite():
    optionflags = doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS
    return doctest.DocTestSuite(optionflags=optionflags,
                                setUp=setUp, tearDown=tearDown)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
group views.
"""
import zc.table.table
import zc.table.column
from zope.app.dependable.interfaces import IDependable
from zope.cachedescriptors.property import Lazy
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.intid.interfaces import IIntIds
from zope.traversing.browser.interfaces import IAbsoluteURL
from zope.interface import implements, directlyProvides
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.publisher.browser import BrowserView
from zope.component import adapts
from zope.component import getUtility
from zope.component import getMultiAdapter
from zope.security.checker import canAccess
from zope.i18n.interfaces.locales import ICollator
from zope.viewlet.viewlet import ViewletBase
from zope.container.interfaces import INameChooser
from zope.security import checkPermission
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.i18n import translate
from z3c.form import field, button, form
from z3c.form.interfaces import HIDDEN_MODE
from zc.table.interfaces import ISortableColumn
from zope.security.interfaces import Unauthorized

from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.skin.containers import TableContainerView
from schooltool.app.browser.app import BaseAddView, BaseEditView
from schooltool.app.browser.app import ContentTitle
from schooltool.app.browser.app import EditRelationships
from schooltool.app.browser.app import RelationshipAddTableMixin
from schooltool.app.browser.app import RelationshipRemoveTableMixin
from schooltool.app.browser.app import RelationshipViewBase
from schooltool.person.interfaces import IPerson
from schooltool.person.interfaces import IPersonFactory
from schooltool.person.browser.person import PersonTableFilter
from schooltool.basicperson.browser.person import BasicPersonTable
from schooltool.basicperson.browser.person import EditPersonRelationships
from schooltool.course.interfaces import ISection
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.schoolyear.interfaces import ISchoolYearContainer
from schooltool.group.group import Group
from schooltool.group.interfaces import IGroup
from schooltool.group.interfaces import IGroupMember
from schooltool.group.interfaces import IGroupContainer, IGroupContained
from schooltool.skin.flourish.viewlet import Viewlet
from schooltool.common.inlinept import InheritTemplate
from schooltool.common.inlinept import InlineViewPageTemplate
from schooltool.skin import flourish
from schooltool import table

from schooltool.common import SchoolToolMessage as _
from schooltool.basicperson.browser.person import FlourishPersonIDCardsViewBase
from schooltool.report.report import OldReportTask
from schooltool.report.browser.report import RequestRemoteReportDialog


class GroupContainerAbsoluteURLAdapter(BrowserView):

    adapts(IGroupContainer, IBrowserRequest)
    implements(IAbsoluteURL)

    def __str__(self):
        container_id = int(self.context.__name__)
        int_ids = getUtility(IIntIds)
        container = int_ids.getObject(container_id)
        url = str(getMultiAdapter((container, self.request), name='absolute_url'))
        return url + '/groups'

    __call__ = __str__


class GroupContainerView(TableContainerView):
    """A Group Container view."""

    __used_for__ = IGroupContainer

    index_title = _("Group index")


class GroupListView(RelationshipViewBase):
    """View for managing groups that a person or a resource belongs to."""

    __used_for__ = IGroupMember

    @property
    def title(self):
        return _("Groups of ${person}", mapping={'person': self.context.title})
    current_title = _("Current Groups")
    available_title = _("Available Groups")

    def getSelectedItems(self):
        """Return a list of groups the current user is a member of."""
        return [group for group in self.context.groups
                if not ISection.providedBy(group)]

    def getAvailableItemsContainer(self):
        app = ISchoolToolApplication(None)
        groups = IGroupContainer(app, {})
        return groups

    def getCollection(self):
        return self.context.groups


class GroupView(BrowserView):
    """A Group info view."""

    __used_for__ = IGroupContained

    def renderPersonTable(self):
        persons = ISchoolToolApplication(None)['persons']
        formatter = getMultiAdapter((persons, self.request),
                                    table.interfaces.ITableFormatter)
        formatter.setUp(table_formatter=zc.table.table.StandaloneFullFormatter,
                        items=self.getPersons(),
                        batch_size=0)
        return formatter.render()

    def getPersons(self):
        return [member for member in self.context.members
                if canAccess(member, 'title')]

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')


class MemberViewPersons(RelationshipViewBase):
    """View class for adding / removing members to / from a group."""

    __used_for__ = IGroupContained

    @property
    def title(self):
        return _("Members of ${group}", mapping={'group': self.context.title})
    current_title = _("Current Members")
    available_title = _("Add Members")

    def getSelectedItems(self):
        """Return a list of current group memebers."""
        return filter(IPerson.providedBy, self.context.members)

    def getAvailableItemsContainer(self):
        return ISchoolToolApplication(None)['persons']

    def getCollection(self):
        return self.context.members


class GroupAddView(BaseAddView):
    """A view for adding a group."""


class GroupEditView(BaseEditView):
    """A view for editing group info."""

    __used_for__ = IGroupContained


class GroupsViewlet(ViewletBase):
    """A viewlet showing the groups a person is in."""

    def update(self):
        self.collator = ICollator(self.request.locale)
        groups = [
            group for group in self.context.groups
            if (canAccess(group, 'title') and
                not ISection.providedBy(group))]

        schoolyears_data = {}
        for group in groups:
            sy = ISchoolYear(group.__parent__)
            if sy not in schoolyears_data:
                schoolyears_data[sy] = []
            schoolyears_data[sy].append(group)

        self.schoolyears = []
        for sy in sorted(schoolyears_data, key=lambda x:x.first, reverse=True):
            sy_info = {'obj': sy,
                       'groups': sorted(schoolyears_data[sy],
                                        cmp=self.collator.cmp,
                                        key=lambda x:x.title)}
            self.schoolyears.append(sy_info)

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')


class FlourishGroupsViewlet(Viewlet):
    """A flourish viewlet showing the groups a person is in."""

    template = ViewPageTemplateFile('templates/f_groupsviewlet.pt')
    render = lambda self, *a, **kw: self.template(*a, **kw)

    def update(self):
        self.collator = ICollator(self.request.locale)
        groups = [
            group for group in self.context.groups
            if (canAccess(group, 'title') and
                not ISection.providedBy(group))]

        schoolyears_data = {}
        for group in groups:
            sy = ISchoolYear(group.__parent__)
            if sy not in schoolyears_data:
                schoolyears_data[sy] = []
            schoolyears_data[sy].append(group)

        self.schoolyears = []
        for sy in sorted(schoolyears_data, key=lambda x:x.first, reverse=True):
            sy_info = {'obj': sy,
                       'groups': sorted(schoolyears_data[sy],
                                        cmp=self.collator.cmp,
                                        key=lambda x:x.title)}
            self.schoolyears.append(sy_info)

    @property
    def canModify(self):
        return canAccess(self.context.__parent__, '__delitem__')


class FlourishGroupFilterWidget(table.table.FilterWidget):

    template = ViewPageTemplateFile('templates/f_group_filter.pt')


class SchoolYearColumn(zc.table.column.GetterColumn):

    def getter(self, item, formatter):
        schoolyear = ISchoolYear(item.__parent__)
        return schoolyear.title

    def getSortKey(self, item, formatter):
        schoolyear = ISchoolYear(item.__parent__)
        return schoolyear.first


class FlourishGroupTableFormatter(table.table.SchoolToolTableFormatter):

    def columns(self):
        title = table.column.LocaleAwareGetterColumn(
            name='title',
            title=_(u"Title"),
            getter=lambda i, f: i.title,
            subsort=True)
        directlyProvides(title, ISortableColumn)
        return [title]

    def makeFormatter(self):
        formatter = table.table.SchoolToolTableFormatter.makeFormatter(self)
        formatter.cssClasses['table'] = 'groups-table relationships-table'
        return formatter


class FlourishGroupListView(EditRelationships):

    current_title = _('Current groups')
    available_title = _('Available groups')

    @Lazy
    def schoolyears(self):
        app = ISchoolToolApplication(None)
        schoolyears = ISchoolYearContainer(app)
        active_schoolyear = schoolyears.getActiveSchoolYear()
        return [schoolyear for schoolyear in schoolyears.values()
                if schoolyear.first >= active_schoolyear.first]

    def getSelectedItems(self):
        return [group for group in self.getCollection()
                if not ISection.providedBy(group) and
                ISchoolYear(group.__parent__) in self.schoolyears]

    def getAvailableItemsContainer(self):
        app = ISchoolToolApplication(None)
        groups = IGroupContainer(app, {})
        return groups

    def getAvailableItems(self):
        result = []
        selected_items = set(self.getSelectedItems())
        for schoolyear in self.schoolyears:
            groups = IGroupContainer(schoolyear)
            result.extend([group for group in groups.values()
                           if group not in selected_items])
        return result

    def getCollection(self):
        return self.context.groups

    def getColumnsAfter(self, prefix):
        columns = super(FlourishGroupListView, self).getColumnsAfter(prefix)
        schoolyear = SchoolYearColumn(
            name='schoolyear',
            title=_(u'School Year'),
            subsort=True)
        directlyProvides(schoolyear, ISortableColumn)
        return [schoolyear] + columns

    def sortOn(self):
        return (('schoolyear', True), ("title", False))

    def setUpTables(self):
        self.available_table = self.createTableFormatter(
            ommit=self.getOmmitedItems(),
            items=self.getAvailableItems(),
            sort_on=self.sortOn(),
            prefix="add_item")
        self.selected_table = self.createTableFormatter(
            filter=lambda l: l,
            items=self.getSelectedItems(),
            sort_on=self.sortOn(),
            prefix="remove_item",
            batch_size=0)

    def getKey(self, item):
        schoolyear = ISchoolYear(item.__parent__)
        return "%s.%s" % (schoolyear.__name__, item.__name__)


class GroupsTertiaryNavigationManager(flourish.page.TertiaryNavigationManager,
                                      ActiveSchoolYearContentMixin):

    template = InlineViewPageTemplate("""
        <ul tal:attributes="class view/list_class">
          <li tal:repeat="item view/items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)

    @property
    def items(self):
        result = []
        active = self.schoolyear
        schoolyears = active.__parent__ if active is not None else {}
        for schoolyear in schoolyears.values():
            url = '%s/%s?schoolyear_id=%s' % (
                absoluteURL(self.context, self.request),
                'groups',
                schoolyear.__name__)
            result.append({
                    'class': schoolyear.first == active.first and 'active' or None,
                    'viewlet': u'<a href="%s">%s</a>' % (url, schoolyear.title),
                    })
        return result


class GroupsAddLinks(flourish.page.RefineLinksViewlet):
    """Manager for Add links in GroupsView"""


class GroupImportLinks(flourish.page.RefineLinksViewlet):
    """Manager for group import links."""


class GroupLinks(flourish.page.RefineLinksViewlet):
    """Manager for public links in GroupView"""

    @property
    def title(self):
        return self.context.title


class GroupAddLinks(flourish.page.RefineLinksViewlet):
    """Manager for Add links in GroupView"""

    def render(self):
        # This check is necessary because the user can be a leader
        # of the context group, which gives him schooltool.edit on it
        if canAccess(self.context.__parent__, '__delitem__'):
            return super(GroupAddLinks, self).render()


class GroupManageActionsLinks(flourish.page.RefineLinksViewlet):
    """Manager for Action links in GroupView"""

    body_template = InlineViewPageTemplate("""
        <ul tal:attributes="class view/list_class">
          <li tal:repeat="item view/renderable_items"
              tal:attributes="class item/class"
              tal:content="structure item/viewlet">
          </li>
        </ul>
    """)

    # We don't want this manager rendered at all
    # if there are no renderable viewlets
    @property
    def renderable_items(self):
        result = []
        for item in self.items:
            render_result = item['viewlet']()
            if render_result and render_result.strip():
                result.append({
                        'class': item['class'],
                        'viewlet': render_result,
                        })
        return result

    def render(self):
        # This check is necessary because the user can be a leader
        # of the context group, which gives him schooltool.edit on it
        if canAccess(self.context.__parent__, '__delitem__'):
            if self.renderable_items:
                return super(GroupManageActionsLinks, self).render()


class GroupDeleteLink(flourish.page.ModalFormLinkViewlet):

    @property
    def dialog_title(self):
        title = _(u'Delete ${group}',
                  mapping={'group': self.context.title})
        return translate(title, context=self.request)

    def render(self, *args, **kw):
        unwrapped = removeSecurityProxy(self.context)
        dependable = IDependable(unwrapped, None)
        if dependable is None or not bool(dependable.dependents()):
            return super(GroupDeleteLink, self).render(*args, **kw)


class GroupAddLinkViewlet(flourish.page.LinkViewlet,
                          ActiveSchoolYearContentMixin):

    @property
    def url(self):
        groups = IGroupContainer(self.schoolyear)
        return '%s/%s' % (absoluteURL(groups, self.request),
                          'addSchoolToolGroup.html')

class GroupAddLinkFromGroupViewlet(GroupAddLinkViewlet):

    @property
    def schoolyear(self):
        return ISchoolYear(self.context.__parent__)

    @property
    def url(self):
        groups = IGroupContainer(self.schoolyear)
        return '%s/%s?camefrom=%s' % (
            absoluteURL(groups, self.request),
            'addSchoolToolGroup.html',
            absoluteURL(self.context, self.request))


class GroupContainerTitle(ContentTitle):

    @property
    def title(self):
        schoolyear = ISchoolYear(self.context)
        return _('Groups for ${schoolyear}',
                 mapping={'schoolyear': schoolyear.title})


class FlourishGroupsView(flourish.page.Page,
                         ActiveSchoolYearContentMixin):

    content_template = InlineViewPageTemplate('''
      <div tal:content="structure context/schooltool:content/ajax/view/container/table" />
    ''')

    @property
    def title(self):
        schoolyear = self.schoolyear
        return _('Groups for ${schoolyear}',
                 mapping={'schoolyear': schoolyear.title})

    @Lazy
    def container(self):
        schoolyear = self.schoolyear
        return IGroupContainer(schoolyear)

    def __call__(self, *args, **kw):
        if not flourish.canView(self.container):
            raise Unauthorized("No permission to view groups.")
        return flourish.page.Page.__call__(self, *args, **kw)


class GroupsTable(table.ajax.Table):

    def columns(self):
        default = table.ajax.Table.columns(self)
        description = zc.table.column.GetterColumn(
            name='description',
            title=_('Description'),
            getter=lambda i, f: i.description or '',
            )
        return default + [description]


class GroupsTableFilter(table.ajax.TableFilter):

    title = _("Group title")


class GroupsTableSchoolYear(flourish.viewlet.Viewlet):

    template = InlineViewPageTemplate('''
      <input type="hidden" name="schoolyear_id"
             tal:define="schoolyear_id view/view/schoolyear/__name__|nothing"
             tal:condition="schoolyear_id"
             tal:attributes="value schoolyear_id" />
    ''')


class GroupsWithSYTable(GroupsTable):

    def columns(self):
        default = table.ajax.Table.columns(self)
        schoolyear = SchoolYearColumn(
            name='schoolyear',
            title=_(u'School Year'),
            subsort=True)
        directlyProvides(schoolyear, ISortableColumn)
        return default + [schoolyear]

    def sortOn(self):
        return (('schoolyear', True), ("title", False))


class GroupListAddRelationshipTable(RelationshipAddTableMixin,
                                    GroupsWithSYTable):

    def updateFormatter(self):
        ommit = self.view.getOmmitedItems()
        available = self.view.getAvailableItems()
        columns = self.columns()
        self.setUp(formatters=[table.table.url_cell_formatter],
                   columns=columns,
                   ommit=ommit,
                   items=available,
                   table_formatter=self.table_formatter,
                   batch_size=self.batch_size,
                   prefix=self.__name__,
                   css_classes={'table': 'data relationships-table'})


class GroupListRemoveRelationshipTable(RelationshipRemoveTableMixin,
                                       GroupsWithSYTable):
    pass


class FlourishGroupContainerDeleteView(flourish.containers.ContainerDeleteView):

    def nextURL(self):
        if 'CONFIRM' in self.request:
            schoolyear = ISchoolYear(self.context)
            url = '%s/%s?schoolyear_id=%s' % (
                absoluteURL(ISchoolToolApplication(None), self.request),
                'groups',
                schoolyear.__name__)
            return url
        return flourish.containers.ContainerDeleteView.nextURL(self)


class FlourishGroupView(flourish.form.DisplayForm, ActiveSchoolYearContentMixin):

    template = InheritTemplate(flourish.page.Page.template)
    content_template = ViewPageTemplateFile('templates/f_group_view.pt')
    fields = field.Fields(IGroup)
    fields = fields.select('title', 'description')

    @property
    def canModify(self):
        return checkPermission('schooltool.edit', self.context)

    @property
    def schoolyear(self):
        return ISchoolYear(self.context.__parent__)

    @property
    def title(self):
        return _('Groups for ${schoolyear}',
                 mapping={'schoolyear': self.schoolyear.title})

    @property
    def subtitle(self):
        return self.context.title

    def done_link(self):
        url = self.request.get('done_link', None)
        if url is not None:
            return url
        app = ISchoolToolApplication(None)
        return self.url_with_schoolyear_id(app, view_name='groups')

    def updateWidgets(self):
        super(FlourishGroupView, self).updateWidgets()
        for widget in self.widgets.values():
            if not widget.value:
                widget.mode = HIDDEN_MODE

    def has_members(self):
        return bool(list(self.context.members))

    def has_leaders(self):
        return bool(list(self.context.leaders))


class FlourishGroupAddView(flourish.form.AddForm, ActiveSchoolYearContentMixin):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('Group Information')
    fields = field.Fields(IGroup)
    fields = fields.select('title', 'description')

    def updateActions(self):
        super(FlourishGroupAddView, self).updateActions()
        self.actions['add'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')

    @button.buttonAndHandler(_('Submit'), name='add')
    def handleAdd(self, action):
        super(FlourishGroupAddView, self).handleAdd.func(self, action)

    @button.buttonAndHandler(_('Cancel'))
    def handle_cancel_action(self, action):
        if 'camefrom' in self.request:
            url = self.request['camefrom']
            self.request.response.redirect(url)
            return
        app = ISchoolToolApplication(None)
        url = self.url_with_schoolyear_id(app, view_name='groups')
        self.request.response.redirect(url)

    def create(self, data):
        group = Group(data['title'], data.get('description'))
        form.applyChanges(self, group, data)
        return group

    def add(self, group):
        chooser = INameChooser(self.context)
        name = chooser.chooseName(u'', group)
        self.context[name] = group
        self._group = group
        return group

    def nextURL(self):
        return absoluteURL(self._group, self.request)

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    @property
    def title(self):
        return _('Groups for ${schoolyear}',
                 mapping={'schoolyear': self.schoolyear.title})


class FlourishGroupEditView(flourish.form.Form, form.EditForm):

    template = InheritTemplate(flourish.page.Page.template)
    label = None
    legend = _('Group Information')
    fields = field.Fields(IGroup)
    fields = fields.select('title', 'description')

    @property
    def title(self):
        return self.context.title

    def update(self):
        return form.EditForm.update(self)

    @button.buttonAndHandler(_('Submit'), name='apply')
    def handleApply(self, action):
        super(FlourishGroupEditView, self).handleApply.func(self, action)
        # XXX: hacky sucessful submit check
        if (self.status == self.successMessage or
            self.status == self.noChangesMessage):
            url = absoluteURL(self.context, self.request)
            self.request.response.redirect(url)

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        url = absoluteURL(self.context, self.request)
        self.request.response.redirect(url)

    def updateActions(self):
        super(FlourishGroupEditView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class FlourishGroupDeleteView(flourish.form.DialogForm, form.EditForm):
    """View used for confirming deletion of a group."""

    dialog_submit_actions = ('apply',)
    dialog_close_actions = ('cancel',)
    label = None

    @button.buttonAndHandler(_("Delete"), name='apply')
    def handleDelete(self, action):
        url = '%s/delete.html?delete.%s&CONFIRM' % (
            absoluteURL(self.context.__parent__, self.request),
            self.context.__name__)
        self.request.response.redirect(url)
        # We never have errors, so just close the dialog.
        self.ajax_settings['dialog'] = 'close'

    @button.buttonAndHandler(_("Cancel"))
    def handle_cancel_action(self, action):
        pass

    def updateActions(self):
        super(FlourishGroupDeleteView, self).updateActions()
        self.actions['apply'].addClass('button-ok')
        self.actions['cancel'].addClass('button-cancel')


class GroupMembersTable(BasicPersonTable):

    prefix = "members"

    def items(self):
        return self.indexItems(self.context.members)


class GroupLeadersTable(BasicPersonTable):

    prefix = "leaders"

    def items(self):
        return self.indexItems(self.context.leaders)


class FlourishMemberViewPersons(EditPersonRelationships):
    """View class for adding / removing members to / from a group."""

    @property
    def title(self):
        return self.context.title

    current_title = _("Current Members")
    available_title = _("Add Members")

    def getSelectedItems(self):
        """Return a list of current group memebers."""
        return filter(IPerson.providedBy, self.context.members)

    def getCollection(self):
        return self.context.members


class FlourishManageGroupsOverview(flourish.page.Content,
                                   ActiveSchoolYearContentMixin):

    body_template = ViewPageTemplateFile(
        'templates/f_manage_groups_overview.pt')

    @property
    def groups(self):
        return IGroupContainer(self.schoolyear, None)

    def groups_url(self):
        return self.url_with_schoolyear_id(self.context, view_name='groups')


class FlourishRequestGroupIDCardsView(RequestRemoteReportDialog):

    report_builder = 'group_id_cards.pdf'
    report_task = OldReportTask


class FlourishGroupIDCardsView(FlourishPersonIDCardsViewBase):

    @property
    def title(self):
        return _('ID Cards for Group: ${group}',
                 mapping={'group': self.context.title})

    @property
    def filename(self):
        sy = ISchoolYear(self.context.__parent__)
        return 'id_cards_%s_%s.pdf' % (
            self.context.title,  sy.title)

    def persons(self):
        collator = ICollator(self.request.locale)
        factory = getUtility(IPersonFactory)
        sorting_key = lambda x: factory.getSortingKey(x, collator)
        sorted_persons = sorted(self.context.members, key=sorting_key)
        result = [self.getPersonData(person)
                  for person in sorted_persons]
        return result


def done_link_url_cell_formatter(group):
    def cell_formatter(value, item, formatter):
        group_url = absoluteURL(group, formatter.request)
        url = '%s?done_link=%s' % (absoluteURL(item, formatter.request),
                                   group_url)
        return '<a href="%s">%s</a>' % (url, value)
    return cell_formatter


class GroupAwarePersonTable(BasicPersonTable):

    def updateFormatter(self):
        group = self.view.context
        self.setUp(formatters=[done_link_url_cell_formatter(group),
                               done_link_url_cell_formatter(group)],
                   table_formatter=self.table_formatter,
                   batch_size=self.batch_size,
                   prefix=self.__name__,
                   css_classes={'table': 'data'})

    def items(self):
        return self.indexItems(self.context)


class GroupAwarePersonTableFilter(PersonTableFilter):

    template = ViewPageTemplateFile('templates/f_group_aware_person_table_filter.pt')


class PersonGroupsTable(table.table.TableContent):

    group_by_column = 'schoolyear'

    @Lazy
    def source(self):
        int_ids = getUtility(IIntIds)
        source = dict([
                ('%s-%d' % (group.__name__, int_ids.getId(group)), group)
                for group in self.context.groups
                ])
        return source

    def columns(self):
        default = table.table.TableContent.columns(self)
        description = zc.table.column.GetterColumn(
            name='description',
            title=_('Description'),
            getter=lambda i, f: i.description or '',
            )
        schoolyear = SchoolYearColumn(
            name='schoolyear',
            title=_(u'School Year'),
            subsort=True)
        directlyProvides(schoolyear, ISortableColumn)
        return default + [description, schoolyear]


class PersonProfileGroupsPart(table.pdf.RMLTablePart):

    table_name = "groups_table"
    title = _("Group memberships")




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool flourish zcml directives.
"""

import zope.component.zcml
import zope.configuration.fields
import zc.resourcelibrary.zcml
from zope.interface import Interface

from schooltool.skin.flourish.resource import patch_zc_resourcelibrary
from schooltool.skin.flourish.resource import IResourceLibrary
from schooltool.skin.flourish import IFlourishLayer


class IDynamicLibraryDirective(Interface):

    class_ = zope.configuration.fields.GlobalObject(
        title=u"Class",
        description=u"A class that selects dependencies and includes.",
        required=True,
        )


IDynamicLibraryDirective.setTaggedValue('keyword_arguments', True)


class ResourceLibrary(zc.resourcelibrary.zcml.ResourceLibrary):

    def library(self, _context, class_, **kwargs):
        patch_zc_resourcelibrary()
        class_dict = dict(kwargs)
        class_dict['__name__'] = self.name
        class_ = type(class_.__name__, (class_, ), class_dict)
        class_.configure()

        _context.action(
            discriminator=('schooltool.skin.flourish.resource-library',
                           self.layer, self.name),
            callable=zope.component.zcml.handler,
            args=('registerAdapter',
                  class_,
                  (self.layer, ),
                  IResourceLibrary,
                  self.name,
                  _context.info),
            )



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Base code for table rendering and filtering.
"""
import pytz
import urllib

import zope.security
from zope.interface import implements
from zope.interface import directlyProvides
from zope.i18n import translate
from zope.browserpage import ViewPageTemplateFile
from zope.cachedescriptors.property import Lazy
from zope.component import queryMultiAdapter
from zope.traversing.browser.absoluteurl import absoluteURL

import zc.table
import zc.table.table
from zc.table.interfaces import IColumnSortedItems
from zc.table.interfaces import ISortableColumn
from zc.table.column import GetterColumn

from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.common import stupid_form_key
from schooltool.skin import flourish
from schooltool.table.batch import Batch
from schooltool.table.interfaces import IFilterWidget
from schooltool.table.interfaces import ITableFormatter

# BBB: imports
from schooltool.common import simple_form_key
from schooltool.table.column import (
    getResourceURL,
    CheckboxColumn, DependableCheckboxColumn,
    DateColumn, LocaleAwareGetterColumn,
    ImageInputColumn, ImageInputValueColumn,
    )

from schooltool.common import SchoolToolMessage as _


class SortUIHeaderMixin(object):
    """Mixin for zc table formatters."""

    def _getColumnSortClass(self, column):
        if not IColumnSortedItems.providedBy(self.items):
            return ""
        col_name = column.name
        for n, (name, reversed) in enumerate(self.items.sort_on):
            if name == col_name:
                if n == 0:
                    return (reversed and "zc-table-sort-desc-primary" or
                                         "zc-table-sort-asc-primary")
                return (reversed and "zc-table-sort-desc" or
                                     "zc-table-sort-asc")
        return "zc-table-sort-asc"

    def _addSortUi(self, header, column):
        css_class = "zc-table-sortable "
        css_class += self._getColumnSortClass(column)
        columnName = column.name
        sort_on_name = zc.table.table.getSortOnName(self.prefix)
        script_name = self.script_name
        return self._header_template(locals())

    def _header_template(self, options):
        options = dict(options)
        template = """
            <span class="%(css_class)s"
                  onclick="javascript: %(script_name)s(
                        '%(columnName)s', '%(sort_on_name)s')">
                %(header)s</span>
            """
        return template % options


class HeaderFormatterMixin(object):

    group_by_column = None
    row = 0

    def __init__(self, context, request, *args, **kw):
        self.context = context
        self.request = request
        self.group_by_column = kw.pop('group_by_column', None)
        kw = self.updateSortParameters(context, request, *args, **kw)
        super(HeaderFormatterMixin, self).__init__(
            context, request, *args, **kw)

    def updateSortParameters(self, *args, **kw):
        fixed_sort_on = self.fixed_sort_on
        if fixed_sort_on:
            if kw.get('ignore_request'):
                kw['sort_on'] = self.getFixedSortOrder(
                    fixed_sort_on, kw.get('sort_on'))
            else:
                kw['sort_on'] = self.getRequestSortOn(
                    fixed_sort_on or (),
                    kw.get('sort_on') or (),
                    prefix=kw.get('prefix'))
                kw['ignore_request'] = True
        return kw

    @property
    def fixed_sort_on(self):
        """Always presort items with this fixed part."""
        if not self.group_by_column:
            return None
        return ((self.group_by_column, False), )

    def getFixedSortOrder(self, fixed_sort_on, sort_on):
        fixed_sort_on = tuple(fixed_sort_on or ())
        fixed = set([so[0] for so in fixed_sort_on])
        also_sort_on = tuple([
            so for so in (sort_on or ())
            if so[0] not in fixed
            ])
        return fixed_sort_on + also_sort_on

    def getRequestSortOn(self, fixed_sort_on, default_sort_on, prefix=None):
        sort_on_name = zc.table.table.getSortOnName(prefix)
        request_sort_on = zc.table.table.getRequestSortOn(
            self.request, sort_on_name) or default_sort_on
        sort_on = self.getFixedSortOrder(fixed_sort_on, request_sort_on)
        return sort_on

    def hasSubGroups(self):
        return (self.group_by_column is not None and
                self.group_by_column in self.columns_by_name)

    def getSubGroup(self, item):
        if not self.hasSubGroups():
            return None
        column = self.columns_by_name[self.group_by_column]
        return column.renderCell(item, self)

    def makeCSSClass(self, cls, column_name):
        cls = cls and cls.strip() or ''
        col_cls = column_name and column_name.replace('_', '-').strip() or ''
        if cls and col_cls:
            cls += ' '
        return cls + col_cls

    def renderHeaders(self):
        result = []
        old_css_class = self.cssClasses.get('th')
        for col in self.visible_columns:
            if self.group_by_column == col.name:
                continue
            self.cssClasses['th'] = self.makeCSSClass(old_css_class, col.name)
            result.append(self.renderHeader(col))
        self.cssClasses['th'] = old_css_class
        return ''.join(result)

    def renderRows(self):
        self.row = 0 # Hack to behave like AlternatingRowFormatterMixin

        if not self.hasSubGroups():
            return ''.join([self.renderRow(item) for item in self.getItems()])

        current_group = None
        result = []
        for item in self.getItems():
            group = self.getSubGroup(item)
            if group != current_group:
                result.append(self.renderSubHeader(item))
                current_group = group
            result.append(self.renderRow(item))
        return ''.join(result)

    def renderCells(self, item):
        return ''.join(
            [self.renderCell(item, col)
             for col in self.visible_columns
             if col.name != self.group_by_column])

    def renderSubHeader(self, item):
        column = self.columns_by_name[self.group_by_column]
        content = column.renderCell(item, self)
        css_class = self.makeCSSClass(self.cssClasses.get('th', ''), column.name)
        cell = '<tr><th class="%s" colspan="%d">%s</th></tr>' % (
            css_class,
            len(self.visible_columns),
            content)
        return cell


class StandaloneHeaderFormatterMixin(HeaderFormatterMixin):

    def getRequestSortOn(self, fixed_sort_on, default_sort_on, prefix=None):
        fixed_sort_on = self.getFixedSortOrder(fixed_sort_on, default_sort_on)
        result = super(StandaloneHeaderFormatterMixin, self).getRequestSortOn(
            fixed_sort_on, (), prefix=prefix)
        return result


class FormFullFormatter(HeaderFormatterMixin,
                        SortUIHeaderMixin,
                        zc.table.table.FormFullFormatter):
    pass


class FilterWidget(object):
    """A simple one field search widget.

    Filters out items in the container by their title.
    """
    implements(IFilterWidget)

    template = ViewPageTemplateFile('templates/filter.pt')

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @property
    def source(self):
        return self.context

    def render(self):
        return self.template()

    def filter(self, list):
        if 'SEARCH' in self.request and 'CLEAR_SEARCH' not in self.request:
            searchstr = self.request['SEARCH'].lower()
            results = [item for item in list
                       if searchstr in item.title.lower()]
        else:
            self.request.form['SEARCH'] = ''
            results = list

        return results

    def active(self):
        return 'SEARCH' in self.request

    def extra_url(self):
        if 'SEARCH' in self.request:
            return '&SEARCH=%s' % urllib.quote(unicode(self.request.get('SEARCH')).encode('UTF-8'))
        return ''


def label_cell_formatter_factory(prefix="", id_getter=None):
    if id_getter is None:
        id_getter = stupid_form_key
    def label_cell_formatter(value, item, formatter):
        return '<label for="%s">%s</label>' % (
            ".".join(filter(None, [prefix, id_getter(item)])), value)
    return label_cell_formatter


def url_cell_formatter(value, item, formatter):
    url = absoluteURL(item, formatter.request)
    return '<a href="%s">%s</a>' % (url, value)


def translate_cell_formatter(value, item, formatter):
    if not value:
        return value
    return translate(value, formatter.request)


def datetime_formatter(value, item, formatter):
    if value is None:
        return ''
    app = ISchoolToolApplication(None)
    preferences = IApplicationPreferences(app)
    preferred_datetime_format = '%s %s' % (preferences.dateformat,
                                           preferences.timeformat)
    app_timezone = pytz.timezone(preferences.timezone)
    local_dt = value.astimezone(app_timezone)
    return local_dt.strftime(preferred_datetime_format)


class NullTableFormatter(object):
    implements(ITableFormatter)

    filter_widget = None
    batch = None

    def __init__(self, context, request):
        self.context, self.request = context, request

    def setUp(self, **kwargs):
        pass

    def makeFormatter(self):
        formatter = zc.table.table.Formatter(
            self.context, self.request, (), columns=())
        return formatter

    def render(self):
        return ""


_undefined = object()


class SchoolToolTableFormatter(object):
    implements(ITableFormatter)

    batch = None
    batch_size = 25
    _items = ()
    prefix = None

    css_classes = None
    _table_formatter = None
    visible_column_names = ()
    group_by_column = None

    def __init__(self, context, request):
        self.context, self.request = context, request

    @property
    def source(self):
        return self.context

    def columns(self):
        title = GetterColumn(name='title',
                             title=_(u"Title"),
                             getter=lambda i, f: i.title,
                             subsort=True)
        directlyProvides(title, ISortableColumn)
        return [title]

    def items(self):
        return self.source.values()

    def ommit(self, items, ommited_items):
        if not ommited_items:
            return items
        ommited_items = set(ommited_items)
        return [item for item in items
                if item not in ommited_items]

    @Lazy
    def batch(self):
        return Batch(self, batch_size=self.batch_size)

    @Lazy
    def filter_widget(self):
        widget = queryMultiAdapter((self.source, self.request),
                                   IFilterWidget)
        return widget

    def filter(self, items):
        # if there is no filter widget, we just return all the items
        if self.filter_widget:
            return self.filter_widget.filter(items)
        else:
            return items

    def sortOn(self):
        return (("title", False),)

    def setUp(self, items=None, ommit=[], filter=None, columns=None,
              columns_before=[], columns_after=[], sort_on=None, prefix="",
              formatters=[], table_formatter=FormFullFormatter,
              batch_size=25, css_classes=None,
              group_by_column=_undefined):

        self.prefix = prefix

        self._table_formatter = table_formatter

        if not columns:
            columns = self.columns()

        if formatters:
            for formatter, column in zip(formatters, columns):
                column.cell_formatter = formatter

        self._columns = columns_before[:] + columns[:] + columns_after[:]

        if items is None:
            items = self.items()

        if not filter:
            filter = self.filter

        self._items = filter(self.ommit(items, ommit))

        if batch_size == 0:
            batch_size = len(list(self._items))

        self.batch_size = batch_size
        self._sort_on = sort_on or self.sortOn() or ()

        if (group_by_column is not _undefined):
            self.group_by_column = group_by_column

        if css_classes:
            self.css_classes = css_classes
        else:
            self.css_classes = {'table': 'data'}

    def extra_url(self):
        extra_url = ""
        if self.filter_widget:
            extra_url += self.filter_widget.extra_url()
        for key, value in self.request.form.items():
            if key.endswith("sort_on"):
                values = [urllib.quote(token) for token in value]
                extra_url += "&%s:tokens=%s" % (key, " ".join(values))
        return extra_url

    def makeFormatter(self):
        if self._table_formatter is None:
            return None
        formatter = self._table_formatter(
            self.source, self.request, self._items,
            visible_column_names=self.visible_column_names,
            columns=self._columns,
            batch_start=self.batch.start, batch_size=self.batch.size,
            sort_on=self._sort_on,
            prefix=self.prefix,
            group_by_column=self.group_by_column)
        formatter.cssClasses.update(dict(self.css_classes))
        formatter.view = self
        return formatter

    def render(self):
        formatter = self.makeFormatter()
        return formatter() if formatter is not None else ''


class TableContent(flourish.content.ContentProvider, SchoolToolTableFormatter):

    def __init__(self, context, request, view):
        flourish.content.ContentProvider.__init__(
            self, context, request, view)

    @property
    def source(self):
        return self.context

    def columns(self):
        title = GetterColumn(name='title',
                             title=_(u"Title"),
                             cell_formatter=url_cell_formatter,
                             getter=lambda i, f: i.title,
                             subsort=True)
        directlyProvides(title, ISortableColumn)
        return [title]

    def update(self):
        flourish.content.ContentProvider.update(self)
        if self._table_formatter is None:
            self.setUp()

    def makeFormatter(self):
        if self._table_formatter is None:
            return None
        formatter = self._table_formatter(
            self.source, self.request, self._items,
            visible_column_names=self.visible_column_names,
            columns=self._columns,
            batch_start=self.batch.start, batch_size=self.batch.size,
            sort_on=self._sort_on,
            prefix=self.prefix,
            group_by_column=self.group_by_column)
        formatter.cssClasses.update(dict(self.css_classes))
        formatter.view = self
        return formatter

    render = SchoolToolTableFormatter.render



class TableContainerView(flourish.page.Page):
    """A base view for containers that use zc.table to display items."""

    empty_message = _('There are none.')
    content_template = ViewPageTemplateFile('templates/f_table_container.pt')
    done_link = ''

    def __init__(self, context, request):
        self.request = request
        self.context = context

    def getColumnsBefore(self):
        return []

    def getColumnsAfter(self):
        return []

    def setUpTableFormatter(self, formatter):
        columns_before = self.getColumnsBefore()
        columns_after = self.getColumnsAfter()
        formatter.setUp(formatters=[url_cell_formatter],
                        columns_before=columns_before,
                        columns_after=columns_after)

    @property
    def container(self):
        return self.context

    def update(self):
        self.table = queryMultiAdapter((self.container, self.request),
                                       ITableFormatter)
        self.setUpTableFormatter(self.table)

    @property
    def deleteURL(self):
        container_url = absoluteURL(self.container, self.request)
        return '%s/%s' % (container_url, 'delete.html')

    def canModify(self):
        return zope.security.canAccess(self.container, '__delitem__')


class DoNotFilter(flourish.EmptyViewlet):

    def extra_url(self):
        return ''

    def filter(self, list):
        return list



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Unit tests for levels.
"""
import unittest
import doctest

from zope.annotation.interfaces import IAttributeAnnotatable
from zope.app.testing import setup
from zope.interface import implements, Interface
from zope.interface.verify import verifyObject
from zope.location.pickling import LocationCopyHook
from zope.component import provideAdapter
from zope.component.hooks import getSite, setSite
from zope.site import SiteManagerContainer
from zope.site.folder import rootFolder

from schooltool.testing.setup import ZCMLWrapper
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.relationship.tests import setUpRelationships
from schooltool.relationship.relationship import getRelatedObjects
from schooltool.level.level import URILevelCourses, URILevel


class AppStub(dict, SiteManagerContainer):
    implements(ISchoolToolApplication)


class CourseStub(object):
    implements(IAttributeAnnotatable)
    def __init__(self, title):
        self.title = unicode(title)
    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.title)


def provideApplicationStub():
    app = AppStub()
    provideAdapter(
        lambda ignored: app,
        adapts=(None,),
        provides=ISchoolToolApplication)
    return app


def doctest_LevelContainer():
    """Tests for LevelContainer.

        >>> from schooltool.level.interfaces import ILevelContainer
        >>> from schooltool.level.level import LevelContainer

    A simple ordered container of levels.

        >>> level_container = LevelContainer()
        >>> verifyObject(ILevelContainer, level_container)
        True

        >>> level_container['vienas'] = u'One'
        >>> level_container['du'] = u'Two'
        >>> level_container['trys'] = u'Three'

        >>> level_container.keys()
        ['vienas', 'du', 'trys']

        >>> level_container.values()
        [u'One', u'Two', u'Three']

        >>> level_container.updateOrder(['trys', 'du', 'vienas'])

        >>> level_container.values()
        [u'Three', u'Two', u'One']

    """


def doctest_Level():
    """Tests for Level.

        >>> from schooltool.level.interfaces import ILevel
        >>> from schooltool.level.level import Level

    Level is a very simple object.

        >>> level = Level(u'1')
        >>> verifyObject(ILevel, level)
        True

        >>> print level.title
        1

    It has a courses attribute to control which courses are taught this level.

        >>> course1 = CourseStub('C1')
        >>> level.courses.add(course1)

        >>> course2 = CourseStub('C2')
        >>> level.courses.add(course2)

        >>> print sorted(level.courses, key=lambda l: l.title)
        [<CourseStub u'C1'>, <CourseStub u'C2'>]

    The courses attribute actually manages relationships between the level and
    the course objects.

        >>> print ['Level %r' % l.title
        ...     for l in getRelatedObjects(
        ...         course1, URILevel, rel_type=URILevelCourses)]
        ["Level u'1'"]

    """


def doctest_LevelVocabulary():
    """Tests for LevelVocabulary.

    Vocabulary of levels:

        >>> from zope.schema.interfaces import IVocabularyTokenized
        >>> from schooltool.level.level import Level
        >>> from schooltool.level.level import LevelVocabulary

        >>> class ContextStub(object):
        ...     implements(Interface)

        >>> context = ContextStub()
        >>> vocabulary = LevelVocabulary(context)
        >>> verifyObject(IVocabularyTokenized, vocabulary)
        True

    If the ISchoolToolApplication object cannot be adapted to
    ILevelContainer, the vocabulary is empty:

        >>> vocabulary.container
        {}

        >>> len(vocabulary)
        0

        >>> list(vocabulary)
        []

        >>> def expand_term(term):
        ...     return (term.token, term.value, term.title)

        >>> level = Level(u'Basic')
        >>> level.__name__ = 'basic'
        >>> expand_term(vocabulary.getTerm(level))
        ('basic-', <schooltool.level.level.Level ...>, u'Basic')

        >>> vocabulary.getTermByToken('basic-')
        Traceback (most recent call last):
        ...
        LookupError: basic-

        >>> level in vocabulary
        False

    Let's provide the needed adapters.

        >>> from schooltool.level.interfaces import ILevelContainer
        >>> from schooltool.level.level import LevelContainer

        >>> levels = LevelContainer()
        >>> levels['basic'] = Level(u'Basic')
        >>> levels['advanced'] = Level(u'Advanced')

        >>> provideAdapter(
        ...     lambda ignored: levels,
        ...     adapts=(ISchoolToolApplication, ),
        ...     provides=ILevelContainer)

    Now we can use the vocabulary.

        >>> vocabulary = LevelVocabulary(context)
        >>> verifyObject(IVocabularyTokenized, vocabulary)
        True

        >>> vocabulary.container
        <schooltool.level.level.LevelContainer ...>

        >>> len(vocabulary)
        2

        >>> [expand_term(term) for term in vocabulary]
        [('basic-', <schooltool.level.level.Level ...>, u'Basic'),
         ('advanced-', <schooltool.level.level.Level ...>, u'Advanced')]

        >>> expand_term(vocabulary.getTerm(levels['basic']))
        ('basic-', <schooltool.level.level.Level ...>, u'Basic')

        >>> levels['basic'] in vocabulary
        True

        >>> expand_term(vocabulary.getTermByToken('basic-'))
        ('basic-', <schooltool.level.level.Level ...>, u'Basic')

    """


def setUp(test):
    setup.placefulSetUp()
    setUpRelationships()
    provideApplicationStub()


def tearDown(test):
    setup.placefulTearDown()


def provideStubAdapter(factory, adapts=None, provides=None, name=u''):
    sm = getSite().getSiteManager()
    sm.registerAdapter(factory, required=adapts, provided=provides, name=name)


def unregisterStubAdapter(factory, adapts=None, provides=None, name=u''):
    sm = getSite().getSiteManager()
    sm.unregisterAdapter(factory, required=adapts, provided=provides, name=name)


def setUpIntegration(test):
    setup.placefulSetUp()
    # Workaround: _clear actually sets the Zope's vocabulary registry and
    #             is called on zope.app.schema.vocabularies import (during
    #             zcml parsing, for example).  When running multiple tests
    #             this ingenious idea fails, so we call it manually.
    from zope.app.schema import vocabulary
    vocabulary._clear()

    zcml = ZCMLWrapper()
    zcml.setUp(
        namespaces={"": "http://namespaces.zope.org/zope"},
        i18n_domain='schooltool')
    zcml.include('schooltool.common', file='zcmlfiles.zcml')
    # We define the default pemissions here, because though widely used,
    # they are currently mangled with other stuff in schooltool.common
    zcml.string('''
      <permission id="schooltool.view" title="View" />
      <permission id="schooltool.edit" title="Edit Info" />
    ''')

    zcml.include('zope.intid')
    zcml.string('''
      <utility
        factory="zope.intid.IntIds"
        provides="zope.intid.interfaces.IIntIds"
      />
      <adapter
          for="persistent.interfaces.IPersistent"
          factory="schooltool.testing.stubs.KeyReferenceStub"
          trusted="y"
          />
    ''')

    zcml.include('schooltool.level', file='level.zcml')
    zcml.include('schooltool.schoolyear', file='schoolyear.zcml')
    zcml.include('schooltool.relationship', file='relationship.zcml')
    provideAdapter(LocationCopyHook)

    root = rootFolder()
    root['app'] = provideApplicationStub()
    setup.createSiteManager(root['app'], setsite=True)
    test.globs.update({
        'zcml': zcml,
        'CourseStub': CourseStub,
        'getRelatedObjects': getRelatedObjects,
        'provideAdapter': provideStubAdapter,
        'unregisterAdapter': unregisterStubAdapter,
        })


def test_suite():
    optionflags = (doctest.NORMALIZE_WHITESPACE |
                   doctest.ELLIPSIS |
                   doctest.REPORT_NDIFF)
    return unittest.TestSuite([
        doctest.DocTestSuite(setUp=setUp, tearDown=tearDown,
                             optionflags=optionflags),
        doctest.DocFileSuite(
            'level-integration.txt',
            setUp=setUpIntegration, tearDown=tearDown,
            optionflags=optionflags),
           ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')





#
#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2009 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Contact information of a Person.

Glue between schooltool.basicperson and schooltool.contact.
"""
import urllib

from zope.schema import getFieldsInOrder
from zope.security.proxy import removeSecurityProxy
from zope.security import checkPermission
from zope.publisher.browser import BrowserView
from zope.browserpage.viewpagetemplatefile import ViewPageTemplateFile
from zope.traversing.browser.absoluteurl import absoluteURL

from z3c.form import field

from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.contact.basicperson import IBoundContact
from schooltool.contact.interfaces import IContactInformation
from schooltool.contact.interfaces import IContact
from schooltool.contact.interfaces import IContactable
from schooltool.contact.interfaces import IContactPerson
from schooltool.contact.interfaces import IAddress, IEmails, IPhones, ILanguages
from schooltool.contact.browser.contact import ContactEditView
from schooltool.contact.browser.contact import FlourishContactEditView
from schooltool.contact.browser.contact import FlourishContactDetails
from schooltool.contact.browser.relationship import get_relationship_title
from schooltool.skin.flourish.page import Page
from schooltool.common import SchoolToolMessage as _


class ContactOverviewView(BrowserView):
    """View class for listing all relevant contact information of a person."""

    __used_for__ = IBoundContact

    @property
    def person(self):
        return self.context.__parent__

    def _extract_attrs(self, contact, interface,
                       conjunctive=", ", add_title=False):
        parts = []
        for name, field in getFieldsInOrder(interface):
            part = getattr(contact, name, None)
            part = part and part.strip() or part
            if not part:
                continue
            if add_title:
                parts.append("%s (%s)" % (part, field.title))
            else:
                parts.append(part)
        return parts

    def buildInfo(self, contact):
        return {
            'link': absoluteURL(contact, self.request),
            'relationship': get_relationship_title(self.person, contact),
            'name': " ".join(self._extract_attrs(
                contact, IContactPerson)),
            'address': ", ".join(self._extract_attrs(contact, IAddress)),
            'emails': ", ".join(self._extract_attrs(contact, IEmails)),
            'phones': list(self._extract_attrs(contact, IPhones, add_title=True)),
            'languages': ", ".join(self._extract_attrs(contact, ILanguages)),
            }

    def getContacts(self):
        contacts = IContactable(removeSecurityProxy(self.person)).contacts
        return [self.buildInfo(contact) for contact in contacts]

    def getRelationships(self):
        bound = IContact(self.person)
        return [relationship_info.extra_info
                for relationship_info in bound.persons.relationships]

    def getPerson(self):
        bound = IContact(self.person)
        return self.buildInfo(bound)


class BoundContactEditView(ContactEditView):
    """Edit form for a bound contact."""
    fields = field.Fields(IContactInformation).omit('photo')


class FlourishBoundContactView(Page):

    pass


class FlourishBoundContactDetails(FlourishContactDetails):

    def canModify(self):
        return checkPermission("schooltool.edit", self.context)


class FlourishBoundContactEditView(FlourishContactEditView):

    fields = field.Fields(IContactInformation).omit('photo')

    def buildFieldsetGroups(self):
        self.fieldset_groups = {
            'address': (
                _('Address'),
                ['address_line_1', 'address_line_2', 'city', 'state',
                 'country', 'postal_code']),
            'contact_information': (
                _('Contact Information'),
                ['home_phone', 'work_phone', 'mobile_phone', 'email',
                 'other_1', 'other_2',
                 'language']),
            }
        self.fieldset_order = (
            'address', 'contact_information')

    def nextURL(self):
        if 'person_id' in self.request:
            person_id = self.request['person_id']
            app = ISchoolToolApplication(None)
            persons = app['persons']
            if person_id in persons:
                return absoluteURL(persons[person_id], self.request)
        base_url = absoluteURL(self.context.__parent__, self.request)
        return "%s/@@manage_contacts.html?%s" % (
            base_url,
            urllib.urlencode([('SEARCH_TITLE',
                               self.context.last_name.encode("utf-8"))]))


class ManageContactsActionViewlet(object):
    @property
    def link(self):
        base_url = absoluteURL(self.context.__parent__, self.request)
        return "%s/@@manage_contacts.html?%s" % (
            base_url,
            urllib.urlencode([('SEARCH_LAST_NAME',
                               self.context.last_name.encode("utf-8"))]))




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2010 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Synchronisation between timetables and calendars.
"""
import pytz

import zope.lifecycleevent.interfaces
from zope.annotation.interfaces import IAnnotations
from zope.component import adapts, adapter, getUtility
from zope.interface import implements
from zope.intid.interfaces import IIntIds
from zope.security.proxy import removeSecurityProxy

from schooltool.timetable import interfaces
from schooltool.app.cal import CalendarEvent, Calendar
from schooltool.calendar.simple import ImmutableCalendar
from schooltool.schoolyear.subscriber import ObjectEventAdapterSubscriber

SCHEDULE_CALENDAR_KEY = 'schooltool.timetable.app.ScheduleCalendar'


class ScheduleCalendarEvent(CalendarEvent):
    """Period scheduled in a calendar."""
    implements(interfaces.IScheduleCalendarEvent)

    schedule = None # schedule responsible for this event
    period = None
    meeting_id = None

    def __init__(self, *args, **kw):
        self.schedule = kw.pop('schedule', None)
        # XXX: int id would be more appropriate here maybe
        self.period = kw.pop('period', None)
        self.meeting_id = kw.pop('meeting_id', None)
        super(ScheduleCalendarEvent, self).__init__(*args, **kw)
        if self.meeting_id is None:
            self.meeting_id = self.unique_id


class UpdateEventTitles(ObjectEventAdapterSubscriber):
    adapts(zope.lifecycleevent.interfaces.IObjectModifiedEvent,
           interfaces.IHaveSchedule)

    def __call__(self):
        title = getattr(self.object, 'title', u'')
        if not title:
            return
        calendar = interfaces.IScheduleCalendar(self.object, None)
        if calendar is None:
            return
        for event in calendar:
            if (interfaces.IScheduleCalendarEvent.providedBy(event)):
                event.title = title


class ImmutableScheduleCalendar(ImmutableCalendar):
    adapts(interfaces.ISchedule)
    implements(interfaces.IImmutableScheduleCalendar)

    schedule = None

    def __init__(self, schedule):
        self.schedule = schedule
        events = tuple(self.createEvents())
        super(ImmutableScheduleCalendar, self).__init__(events=events)

    def makeGUID(self, date, period, int_ids=None):
        if int_ids is None:
            int_ids = getUtility(IIntIds)
        return u'%s.%s.%s' % (
            date.isoformat(),
            int_ids.getId(self.schedule),
            int_ids.getId(period),
            )

    def createEvents(self):
        int_ids = getUtility(IIntIds)
        owner = interfaces.IHaveSchedule(self.schedule)
        title = getattr(owner, 'title', u'')

        schedule = self.schedule
        if (schedule.first is None or
            schedule.last is None):
            return # Empty schedule

        meetings = schedule.iterMeetings(schedule.first, schedule.last)

        for meeting in meetings:
            # We need to convert dtstart to UTC, because calendar
            # events insist on storing UTC time.
            dtstart = meeting.dtstart.astimezone(pytz.UTC)
            guid = self.makeGUID(dtstart.date(), meeting.period,
                                 int_ids=int_ids)

            event = ScheduleCalendarEvent(
                dtstart, meeting.duration, title,
                schedule=schedule,
                period=meeting.period,
                meeting_id=meeting.meeting_id,
                unique_id=guid)

            yield event


class ScheduleCalendar(Calendar):
    implements(interfaces.IScheduleCalendar)

    _synchronised_attrs = ('dtstart', 'duration',
                           'period', 'meeting_id',
                           'title')

    def updateEvent(self, event, other_event):
        _unspecified = object()
        changed = False
        for attr in self._synchronised_attrs:
            new_val = getattr(other_event, attr, _unspecified)
            if (new_val is not _unspecified and
                new_val != getattr(event, attr)):
                setattr(event, attr, new_val)
                changed = True
        return changed

    def updateSchedule(self, schedule):
        schedule = removeSecurityProxy(schedule)
        schedule_cal = interfaces.IImmutableScheduleCalendar(schedule)
        if schedule_cal is None:
            self.removeSchedule(schedule)
            return

        old_events = dict(
            [(e.unique_id, e) for e in removeSecurityProxy(self)
              if e.schedule is schedule])

        new_events = dict([(e.unique_id, e) for e in schedule_cal])

        old_set = set(old_events)
        new_set = set(new_events)

        for uid in sorted(old_set - new_set):
            self.removeEvent(old_events[uid])

        for uid in sorted(new_set - old_set):
            self.addEvent(new_events[uid])

        for uid in sorted(new_set & old_set):
            self.updateEvent(old_events[uid], new_events[uid])

    def removeSchedule(self, schedule):
        schedule = removeSecurityProxy(schedule)
        old_events = sorted([e for e in removeSecurityProxy(self)])
        for event in old_events:
            self.removeEvent(event)


@adapter(interfaces.IHaveSchedule)
def getScheduleCalendar(owner):
    annotations = IAnnotations(owner)
    try:
        return annotations[SCHEDULE_CALENDAR_KEY]
    except KeyError:
        calendar = ScheduleCalendar(owner)
        annotations[SCHEDULE_CALENDAR_KEY] = calendar
        return calendar
getScheduleCalendar.factory = ScheduleCalendar


class UpdateScheduleCalendar(ObjectEventAdapterSubscriber):
    def __call__(self):
        owner = interfaces.IHaveSchedule(self.object, None)
        if owner is None:
            return
        calendar = interfaces.IScheduleCalendar(owner, None)
        if calendar is None:
            return
        container = interfaces.IScheduleContainer(owner, None)
        if container is None:
            return

        calendar.updateSchedule(container)


class RemoveScheduleCalendar(ObjectEventAdapterSubscriber):
    def __call__(self):
        owner = interfaces.IHaveSchedule(self.object, None)
        if owner is None:
            return
        calendar = interfaces.IScheduleCalendar(owner, None)
        if calendar is None:
            return
        container = interfaces.IScheduleContainer(owner, None)
        if container is None:
            return
        calendar.removeSchedule(self.object)




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Browser views for the SchoolTool application.
"""
import calendar
import itertools

from zope.interface import implements
from zope.publisher.interfaces.browser import IBrowserPage
from zope.component import adapts, queryMultiAdapter
from zope.container.interfaces import IWriteContainer
from zope.size.interfaces import ISized
from zope.traversing.interfaces import IPathAdapter, ITraversable
from zope.security import checkPermission
from zope.security.interfaces import IPrincipal
from zope.authentication.interfaces import IUnauthenticatedPrincipal
from zope.app.dependable.interfaces import IDependable
from zope.tales.interfaces import ITALESFunctionNamespace
from zope.security.proxy import removeSecurityProxy
from zope.security.checker import canAccess, canWrite

from pytz import timezone

from schooltool.app.interfaces import IApplicationPreferences
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.skin.flourish.interfaces import IContentProviders, IPageBase
from schooltool.person.interfaces import IPerson

from schooltool.common import SchoolToolMessage as _


class SchoolToolAPI(object):
    """TALES function namespace for SchoolTool specific actions.

    In a page template you can use it as follows:

        tal:define="app context/schooltool:app"

    """

    implements(IPathAdapter, ITALESFunctionNamespace)

    engine = None

    devmode = False

    def __init__(self, context):
        self.context = context

    def setEngine(self, engine):
        self.engine = engine

    @property
    def app(self):
        """Return the ISchoolToolApplication.

        Sample usage in a page template:

            <a tal:attributes="href context/schooltool:app/@@absolute_url">
               Front page
            </a>

        """
        return ISchoolToolApplication(None)

    @property
    def content(self):
        """Get traversable content providers for the context.

        Say, we have a viewlet manager named ExtraInfo, registered for persons.
        As viewlet managers implement IContentProvider, we can do:

        <div tal:repeat="person view/persons">
          <p tal:content="person/title"></p>
          <div tal:replace="structure person/schooltool:content/ExtraInfo"></div>
        </div>

        """
        if self.engine is None:
            return None
        vars = self.engine.vars
        context = self.context
        request = vars.get('request', None)
        view = vars.get('view', None)
        providers = queryMultiAdapter(
            (context, request, view),
            IContentProviders)
        if ITALESFunctionNamespace.providedBy(providers):
            providers.setEngine(self.engine)
        return providers

    @property
    def page_content(self):
        """Get traversable content providers for the context.

        Say, we have a viewlet manager named ExtraInfo, registered for persons.
        As viewlet managers implement IContentProvider, we can do:

        <div tal:repeat="person view/persons">
          <p tal:content="person/title"></p>
          <div tal:replace="structure person/schooltool:page_content/ExtraInfo"></div>
        </div>

        """
        if self.engine is None:
            return None
        vars = self.engine.vars
        context = self.context
        request = vars.get('request', None)
        page = view = vars.get('view', None)
        while (page is not None and
               not IPageBase.providedBy(page)):
            page = getattr(page, '__parent__', None)
        if page is None:
            page = view
            while (page is not None and
                   not IBrowserPage.providedBy(page)):
                page = getattr(page, '__parent__', None)
        providers = queryMultiAdapter(
            (context, request, page),
            IContentProviders)
        if ITALESFunctionNamespace.providedBy(providers):
            providers.setEngine(self.engine)
        return providers

    @property
    def person(self):
        """Adapt context to IPerson, default to None.

        Sample usage in a page template:

            <a tal:define="person request/principal/schooltool:person"
               tal:condition="person"
               tal:attributes="person/calendar/@@absolute_url">
               My calendar
            </a>

        """
        return IPerson(self.context, None)

    @property
    def authenticated(self):
        """Check whether context is an authenticated principal.

        Sample usage in a page template:

            <tal:span tal:define="user request/principal"
                      tal:condition="user/schooltool:authenticated"
                      tal:replace="user/title">
              User title
            </tal:span>
            <tal:span tal:define="user request/principal"
                      tal:condition="not:user/schooltool:authenticated">
              Anonymous
            </tal:span>

        """
        if self.context is None: # no one is logged in
            return False
        if not IPrincipal.providedBy(self.context):
            raise TypeError("schooltool:authenticated can only be applied"
                            " to a principal but was applied on %r" % self.context)
        return not IUnauthenticatedPrincipal.providedBy(self.context)

    @property
    def preferences(self):
        """Return ApplicationPreferences for the SchoolToolApplication.

        Sample usage in a page template:

          <div tal:define="preferences context/schooltool:preferences">
            <b tal:content="preferences/title"></b>
          </div>

        """
        return IApplicationPreferences(self.app)

    @property
    def has_dependents(self):
        """Check whether an object has dependents (via IDependable).

        Objects that have dependents cannot be removed from the system.

        Sample usage in a page template:

          <input type="checkbox" name="delete"
                 tal:attributes="disabled obj/schooltool:has_dependents" />

        """
        # We cannot adapt security-proxied objects to IDependable.  Unwrapping
        # is safe since we do not modify anything, and the information whether
        # an object can be deleted or not is not classified.
        unwrapped_context = removeSecurityProxy(self.context)
        dependable = IDependable(unwrapped_context, None)
        if dependable is None:
            return False
        else:
            return bool(dependable.dependents())

    def can_view(self):
        return checkPermission("schooltool.view", self.context)

    def can_edit(self):
        return checkPermission("schooltool.edit", self.context)

    def can_delete(self):
        container = self.context.__parent__
        if not IWriteContainer.providedBy(container):
            raise NotImplementedError()
        return canAccess(container, '__delitem__')


class PathAdapterUtil(object):

    adapts(None)
    implements(IPathAdapter, ITraversable)

    def __init__(self, context):
        self.context = context


class SortBy(PathAdapterUtil):
    """TALES path adapter for sorting lists.

    In a page template you can use it as follows:

        tal:repeat="something some_iterable/sortby:attribute_name"

    In Python code you can write

        >>> l = [{'name': 'banana'}, {'name': 'apple'}]
        >>> SortBy(l).traverse('name')
        [{'name': 'apple'}, {'name': 'banana'}]

    You can sort arbitrary iterables, not just lists.  The sort key
    can refer to a dictionary key, or an object attribute.
    """

    def traverse(self, name, furtherPath=()):
        """Return self.context sorted by a given key."""
        # We need to get the first item without losing it forever
        iterable = iter(self.context)
        try:
            first = iterable.next()
        except StopIteration:
            return [] # We got an empty list
        iterable = itertools.chain([first], iterable)
        # removeSecurityProxy is safe here because subsequent getattr() will
        # raise Unauthorized or ForbiddenAttribute as appropriate.  It is
        # necessary here to fix http://issues.schooltool.org/issue174
        if hasattr(removeSecurityProxy(first), name):
            items = [(getattr(item, name), item) for item in iterable]
        else:
            items = [(item[name], item) for item in iterable]
        items.sort()
        return [row[-1] for row in items]


class CanAccess(PathAdapterUtil):
    """TALES path adapter for checking access rights.

    In a page template this adapter can be used like this:

        <p tal:condition="context/can_access:title"
           tal:content="context/title" />

    """

    def traverse(self, name, furtherPath=()):
        """Returns True if self.context.(name) can be accessed."""
        return canAccess(self.context, name)


class CanModify(PathAdapterUtil):
    """TALES path adapter for checking access rights.
    """

    def traverse(self, name, furtherPath=()):
        """Returns True if self.context.(name) can be changed."""
        return canWrite(self.context, name)


class FilterAccessible(PathAdapterUtil):
    """TALES path adapter for XXX

    In a page template this adapter can be used like this:

        <p tal:repeat="group context/groups/filter_accessible:title"
           tal:content="group/title" />

    """

    def traverse(self, name, furtherPath=()):
        """XXX"""
        return [item for item in self.context
                if canAccess(item, name)]


class SortedFilterAccessible(PathAdapterUtil):
    """TALES path adapter for XXX

    In a page template this adapter can be used like this:

        <p tal:repeat="group context/groups/sorted_filter_accessible:title"
           tal:content="group/title" />

    """

    def traverse(self, name, furtherPath=()):
        """XXX"""
        filtered = FilterAccessible(self.context).traverse(name)
        return SortBy(filtered).traverse(name)


class SchoolToolSized(object):
    """An adapter to provide number of persons in a SchoolTool instance."""

    implements(ISized)

    def __init__(self, app):
        self._app = app

    def sizeForSorting(self):
        return (_("Persons"), len(self._app['persons']))

    def sizeForDisplay(self):
        num = self.sizeForSorting()[1]
        if num == 1:
            msgid = _("1 person")
        else:
            msgid = _("${number} persons", mapping={'number': num})
        return msgid


class ViewPreferences(object):
    """Preference class to attach to views."""

    def __init__(self, request):
        try:
            app = ISchoolToolApplication(None)
            prefs = IApplicationPreferences(app)
        except (ValueError, TypeError):
            prefs = None

        if prefs is not None:
            self.dateformat = prefs.dateformat
            self.timeformat = prefs.timeformat
            self.first_day_of_week = prefs.weekstart
            self.timezone = timezone(prefs.timezone)
        else:
            # no user, no application - test environment
            self.dateformat = '%Y-%m-%d'
            self.timeformat = '%H:%M'
            self.first_day_of_week = calendar.MONDAY
            self.timezone = timezone('UTC')

    def renderDatetime(self, dt):
        dt = dt.astimezone(self.timezone)
        return dt.strftime('%s %s' % (self.dateformat, self.timeformat))


def same(obj1, obj2):
    """Return True if the references obj1 and obj2 point to the same object.

    The references may be security-proxied.
    """
    return removeSecurityProxy(obj1) is removeSecurityProxy(obj2)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Widgets.
"""
import zope.schema
import zope.app.form
from zope.traversing.browser.absoluteurl import absoluteURL
from zope.schema.interfaces import IField, IDate
from zope.schema.fieldproperty import FieldProperty
from zope.component import adapts, adapter
from zope.interface import implements, implementsOnly, implementer
from zope.interface import Interface
from zope.html.field import IHtmlFragmentField
from zope.publisher.interfaces.browser import IBrowserRequest
import zope.html.widget

import zc.resourcelibrary
from zc.datetimewidget.datetimewidget import DateWidget

import z3c.form
import z3c.form.interfaces
from z3c.form.widget import Widget, FieldWidget, ComputedWidgetAttribute
from z3c.form.converter import BaseDataConverter
from z3c.form.converter import FormatterValidationError
from z3c.form.browser.text import TextWidget
from z3c.form.browser.textarea import TextAreaWidget

from schooltool.common import parse_date
from schooltool.common import SchoolToolMessage as _
from schooltool.app.interfaces import ISchoolToolApplication


class IDateTextWidget(Interface):
    pass


class CustomDateTextWidget(TextWidget, DateWidget):
    implements(IDateTextWidget)

    def date_selector_button(self):
        real_name = self.name
        self.name = self.id
        result = self._render("")
        self.name = real_name
        return result


def CustomDateFieldTextWidget(field, request):
    """IFieldWidget factory for MyWidget."""
    return FieldWidget(field, CustomDateTextWidget(request))


class CustomDateDataConverter(BaseDataConverter):
    """A special data converter for iso dates."""

    adapts(IDate, CustomDateTextWidget)

    def toWidgetValue(self, value):
        """See interfaces.IDataConverter"""
        if value is self.field.missing_value:
            return u''
        try:
            return value.strftime("%Y-%m-%d")
        except (ValueError,):
            # XXX: may be evil, but this allows users to fix incorrect
            #      dates entered before we added the >= 1900 check
            return str(value)

    def toFieldValue(self, value):
        """See interfaces.IDataConverter"""
        if value == u'':
            return self.field.missing_value
        try:
            value = parse_date(value)
        except (ValueError,):
            raise FormatterValidationError(
                _("The datetime string did not match the pattern yyyy-mm-dd"),
                value)
        try:
            value.strftime("%Y-%m-%d")
        except (ValueError,):
            raise FormatterValidationError(
                _('Year has to be equal or greater than 1900'),
                value)
        return value


class ICkeditorConfig(Interface):

    width = zope.schema.Int(
        title=u"Width",
        description=u"Editor frame width")

    height = zope.schema.Int(
        title=u"Height",
        description=u"Editor frame height")

    toolbar = zope.schema.TextLine(
        title=u"Toolbar configuration",
        description=u"The name of the toolbar configuration to use.")

    path = zope.schema.TextLine(
        title=u"Relative configuration path",
        description=u"Path to the CKEditor configuration javascript file.")


class CkeditorConfig(object):
    """Configuration of the CKEditor widget."""
    implements(ICkeditorConfig)

    toolbar = u"schooltool"
    path = u"/@@/editor_config.js"

    def __init__(self, width=430, height=300):
        self.width = width
        self.height = height

    def __repr__(self):
        return '<%s.%s (%d x %d, %r toolbar, %r)>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            self.width, self.height,
            self.toolbar, self.path)


class ICkeditorWidget(z3c.form.interfaces.IWidget):

    config = zope.schema.Object(
        title=u"Configuration",
        description=u"CKeditor configuration.",
        schema=ICkeditorConfig)


class CkeditorWidgetBase(object):

    config = None # ICkeditorConfig

    @property
    def editor_var_name(self):
        return 'CKeditor_%s' % str(hash(self.name)).replace('-', 'u')

    @property
    def element_id(self):
        return self.id

    @property
    def script(self):
        zc.resourcelibrary.need("ckeditor")
        config = self.config

        app_url = absoluteURL(ISchoolToolApplication(None), self.request)
        config_path = '%s%s' % (app_url, config.path)

        # XXX: using some values that may be not JS safe
        return '''
            <script type="text/javascript" language="JavaScript">
                var %(variable)s = new CKEDITOR.replace("%(id)s",
                    {
                        height: %(height)s,
                        width: %(width)s,
                        customConfig: "%(customConfigPath)s"
                    }
                );
            </script>
            ''' % {
            'id': self.element_id,
            'variable': self.editor_var_name,
            'width': config.width,
            'height': config.height,
            'toolbar': config.toolbar,
            'customConfigPath': config_path,
            }


class CkeditorFormlibWidget(zope.app.form.browser.TextAreaWidget,
                             CkeditorWidgetBase):

    def __init__(self, *args, **kw):
        zope.app.form.browser.TextAreaWidget.__init__(self, *args, **kw)
        self.config = CkeditorConfig()

    @property
    def editor_var_name(self):
        return 'CKeditor_%s' % self.name.split('.', 1)[-1]

    @property
    def element_id(self):
        return self.name

    def __call__(self):
        zc.resourcelibrary.need("ckeditor")
        textarea = zope.app.form.browser.TextAreaWidget.__call__(self)
        script = self.script
        return '%s\n%s' % (textarea, script)


class IHTMLFragmentWidget(zope.interface.Interface):
    """The HTML element 'core' attributes."""

    id = zope.schema.BytesLine(
        title=u'Id',
        description=(u'This attribute assigns a name to an element. This '
                     u'name must be unique in a document.'),
        required=False)


class HTMLFragmentWidget(object):
    implements(IHTMLFragmentWidget)

    id = FieldProperty(IHTMLFragmentWidget['id'])


class CkeditorZ3CFormWidget(TextAreaWidget,
                             HTMLFragmentWidget,
                             CkeditorWidgetBase):
    """CKEditor z3c.form widget implementation."""
    implementsOnly(ICkeditorWidget)

    config = None
    value = u''

    _adapterValueAttributes = Widget._adapterValueAttributes + ('config', )


@adapter(IField, z3c.form.interfaces.IFormLayer)
@implementer(z3c.form.interfaces.IFieldWidget)
def CkeditorFieldWidget(field, request):
    """Editor widget bound to a field."""
    return FieldWidget(field, CkeditorZ3CFormWidget(request))


# The default configuration for CkeditorZ3CFormWidget
Ckeditor_config = ComputedWidgetAttribute(
    lambda a: CkeditorConfig(),
    request=IBrowserRequest,
    widget=ICkeditorWidget,
    )


# XXX: EditFormCkeditorConfig will now be applied to all add and edit forms.
#      This is wrong, but we do not have standard SchoolTool add/edit
#      z3c.forms yet, like we do with formlib; as a result, we do not have
#      standard interfaces (for example ISchooltoolAddForm) that we could
#      hook the config on.
Ckeditor_addform_config = ComputedWidgetAttribute(
    lambda a: CkeditorConfig(306, 200),
    context=None,
    request=IBrowserRequest,
    view=z3c.form.interfaces.IAddForm,
    field=IHtmlFragmentField,
    widget=ICkeditorWidget,
    )

Ckeditor_editform_config = ComputedWidgetAttribute(
    lambda a: CkeditorConfig(306, 200),
    context=None,
    request=IBrowserRequest,
    view=z3c.form.interfaces.IEditForm,
    field=IHtmlFragmentField,
    widget=ICkeditorWidget,
    )




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Course and Section related interfaces
"""

import zope.schema
from zope.container.interfaces import IContainer, IContained
from zope.container.constraints import contains, containers
from zope.interface import Interface, Attribute

from schooltool.group.interfaces import IBaseGroup as IGroup
from schooltool.common import SchoolToolMessage as _


class ICourse(Interface):
    """Courses are similar to groups, membership is restricted to Sections."""

    __name__ = zope.schema.TextLine(
        title=_("SchoolTool ID"),
        description=_(
            """An internal identifier of this course."""),
        required=True)

    title = zope.schema.TextLine(title=_("Title"))

    description = zope.schema.Text(
        title=_("Description"),
        required=False)

    sections = Attribute(
        """The Sections that implement this course material,
           see schooltool.relationship.interfaces.IRelationshipProperty.""")

    course_id = zope.schema.TextLine(
        title=_("Course ID"),
        description=_(
            """School identifier of this course."""),
        required=False)

    government_id = zope.schema.TextLine(
        title=_("Alternate ID"),
        required=False,
        description=_(
            "Additional identifier for outside tracking or other purposes."))

    credits = zope.schema.Decimal(
        title=_("Credits"),
        required=False)


class ICourseContainer(IContainer):
    """Container of Courses."""

    contains(ICourse)


class ICourseContainerContainer(IContainer):
    """Container of Courses."""

    contains(ICourseContainer)


class ICourseContained(ICourse, IContained):
    """Courses contained in an ICourseContainer."""

    containers(ICourseContainer)


class ISectionBase(IGroup):
    """Sections are groups of users in a particular meeting of a Course."""

    label = zope.schema.TextLine(
        title=_("Label"),
        required=False,
        description=_(
            "An identifier for a section, made up of instructor "
            "names, courses, and meeting time."))

    title = zope.schema.TextLine(
        title=_("Title"),
        required=True)

    description = zope.schema.Text(
        title=_("Description"),
        required=False)

    instructors = Attribute(
        """A list of Person objects in the role of instructor""")

    members = Attribute(
        """Students listed in the role of member""")

    courses = Attribute(
        """A list of courses this section is a member of.""")

    size = Attribute(
        """The number of member students in the section.""")

    previous = Attribute(
        """The previous section.""")

    next = Attribute(
        """The next section.""")

    linked_sections = Attribute(
        """Chain of sections linked by previous/next with this one.""")


class ISection(ISectionBase):

    __name__ = zope.schema.TextLine(
        title=_("SchoolTool ID"),
        description=_(
            """An internal identifier of this section."""),
        required=True)


class ISectionContainer(IContainer):
    """A container for Sections."""

    contains(ISection)


class ISectionContainerContainer(IContainer):
    """A container for Section containers."""

    contains(ISectionContainer)


class ISectionContained(ISection, IContained):
    """Sections in a SectionContainer."""

    containers(ISectionContainer)


class ILearner(Interface):

    def sections():
        """List of all the sections this learner belongs to."""


class IInstructor(Interface):

    def sections():
        """List of all the sections this instructor is teaching to."""



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2007 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool basic person interfaces.
"""
from zope.container.interfaces import IContainer
from zope.container.interfaces import IOrderedContainer
from zope.schema import Date, Choice, TextLine, Bool, List, Int
from zope.configuration.fields import PythonIdentifier
from zope.interface import Interface, Attribute

from zope.schema.interfaces import IVocabularyTokenized

from schooltool.app.utils import vocabulary
from schooltool.common import SchoolToolMessage as _


class IBasicPerson(Interface):
    """Marker interface for a basic person."""

    prefix = TextLine(
        title=_(u"Prefix"),
        required=False,
        )

    first_name = TextLine(
        title=_(u"First name"),
        required=True,
        )

    middle_name = TextLine(
        title=_(u"Middle name"),
        required=False,
        )

    last_name = TextLine(
        title=_(u"Last name"),
        required=True,
        )

    suffix = TextLine(
        title=_(u"Suffix"),
        required=False,
        )

    preferred_name = TextLine(
        title=_(u"Preferred name"),
        required=False,
        )

    gender = Choice(
        title=_(u"Gender"),
        vocabulary=vocabulary([('male', _('Male')), ('female', _('Female')),]),
        required=False,
        )

    birth_date = Date(
        title=_(u"Birth date"),
        description=_(u"(yyyy-mm-dd)"),
        required=False,
        )

    advisors = Attribute("""Advisors of the person""")

    advisees = Attribute("""Advisees of the person""")


class IBasicPersonVocabulary(IVocabularyTokenized):
    """Marker interface for vocabularies that list basic persons."""


# XXX should be in skin or common, or more properly - core
class IGroupVocabulary(IVocabularyTokenized):
    """Marker interface for vocabularies that list schooltool groups."""


class IAdvisor(Interface):

    students = Attribute("""Students being advised by the advisor.""")


    def addStudent(student):
        """Add a student to the advised students list."""

    def removeStudent(student):
        """Remove this student from the advised students list."""


class IDemographics(IContainer):
    """Demographics data storage for a person.

    Stores any kind of data defined by field descriptions that are set
    up for the person container.
    """


class IDemographicsFields(IOrderedContainer):
    """Demographics field storage."""

    def filter_key(key):
        """Return the subset of fields whose limited_keys list is either
           empty, or it contains the key passed"""

    def filter_keys(keys):
        """Return the subset of fields whose limited_keys list is either
           empty, or it contains one of the keys passed"""


class FilterKeyList(List):
    """Marker field to pin widgets on."""


class IFieldDescription(Interface):
    """Demographics field."""

    title = TextLine(
        title = _(u"Title"),
        description = _(u"As it should appear on forms and reports."))

    name = PythonIdentifier(
        title = _(u"ID"),
        description = _(u"A unique one word alphanumeric identifier."))

    required = Bool(
        title = _(u"Required"))

    limit_keys = FilterKeyList(
        title = _(u"Limit to group(s)"),
        description = _(u"If you select one or more groups below, this field "
                         "will only be displayed in forms and reports for "
                         "members of the selected groups."),
        value_type=Choice(
            source="schooltool.basicperson.limit_keys_vocabulary",
            required=True,
            ),
        required=False)


class EnumValueList(List):
    """Marker field to pin custom validation on."""


class IEnumFieldDescription(IFieldDescription):
    """Enumeration demographics field."""

    items = EnumValueList(
        title = _('Selection list'),
        description = _(u"Enter the valid values for the field below.  One "
                         "value per line.  These values will be displayed "
                         "as a menu in forms."))


class IIntFieldDescription(IFieldDescription):
    """Integer demographics field."""

    min_value = Int(
        title=_(u'Minimum value'),
        required=False)

    max_value = Int(
        title=_(u'Maximum value'),
        required=False)


class IFieldFilterVocabulary(IVocabularyTokenized):
    """Marker interface for vocabularies that give keys that are used
    to filter demographics fields for the context.
    """


class IAddEditViewTitle(Interface):
    """Demographics field add/edit view title."""


class ILimitKeysLabel(Interface):
    """Demographics field add/edit view limit keys label text."""


class ILimitKeysHint(Interface):
    """Demographics field add/edit view limit keys hint text."""




#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2005 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
SchoolTool timetable csv import.

XXX: This should be in the schooltool.timetable package.
"""
import csv
import pprint

from zope.container.interfaces import INameChooser
from zope.proxy import sameProxiedObjects
from zope.traversing.browser.absoluteurl import absoluteURL

import schooltool.skin.flourish.page
from schooltool.app.browser.app import ActiveSchoolYearContentMixin
from schooltool.app.interfaces import ISchoolToolApplication
from schooltool.app.browser.csvimport import BaseCSVImportView
from schooltool.course.interfaces import ICourseContainer
from schooltool.course.interfaces import ISectionContainer
from schooltool.course.section import Section
from schooltool.app.browser.csvimport import FlourishBaseCSVImportView
from schooltool.schoolyear.interfaces import ISchoolYear
from schooltool.term.interfaces import ITermContainer
from schooltool.term.term import getNextTerm
from schooltool.timetable.interfaces import ITimetableContainer
from schooltool.timetable.interfaces import IScheduleContainer
from schooltool.timetable.timetable import SelectedPeriodsSchedule
from schooltool.skin import flourish

from schooltool.common import SchoolToolMessage as _


class TimetableImportErrorCollection(object):

    def __init__(self):
        self.generic = []
        self.day_ids = []
        self.periods = []
        self.persons = []
        self.courses = []
        self.sections = []
        self.records = []

    def anyErrors(self):
        return bool(self.generic or self.day_ids or self.periods
                    or self.persons or self.courses or self.sections
                    or self.records)

    def __str__(self):
        return '%s:\n%s' % (
            self.__class__.__name__,
            pprint.pformat(self.__dict__))

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.__dict__)


class TimetableCSVImporter(object):
    """A timetable CSV parser and importer.

    You will most likely want to use the importFromCSV(csvdata) method.

    This class does not use exceptions for handling errors excessively
    because of the nature of error-checking: we want to gather many errors
    in one sweep and present them to the user at once.
    """

    app = None
    schoolyear = None
    persons = None
    errors = None
    charset = None

    terms = None

    def __init__(self, schoolyear, charset=None):
        self.app = ISchoolToolApplication(None)
        self.schoolyear = schoolyear
        self.persons = self.app['persons']
        self.errors = TimetableImportErrorCollection()
        self.charset = charset

    def importFromCSV(self, sections_csv):
        """Import sections from CSV data.

        At the top of the file there should be a header row:

        term
        -or-
        first term, last term

        Then an empty line should follow, and the remaining CSV data should
        consist of chunks like this:

        course_id, instructor_id
        timetable_id
        day_id, period_id
        day_id, period_id
        timetable_id
        day_id, period_id
        ...
        ***
        student_id
        student_id
        ...

        """
        if '\n' not in sections_csv:
            self.errors.generic.append(_("No data provided"))
            return False

        rows = self.parseCSVRows(sections_csv.splitlines())

        if rows[1]:
            self.errors.generic.append(_("Row 2 is not empty"))
            return False

        self.importHeader(rows[0])
        if self.errors.anyErrors():
            return False

        self.importChunks(rows[2:], dry_run=True)
        if self.errors.anyErrors():
            return False

        self.importChunks(rows[2:], dry_run=False)
        if self.errors.anyErrors():
            return False
        return True

    def importChunks(self, rows, dry_run=True):
        """Import chunks separated by empty lines."""
        chunk_start = 0
        for i, row in enumerate(rows):
            if not row:
                if rows[chunk_start]:
                    self.importChunk(rows[chunk_start:i],
                                     chunk_start + 3, dry_run)
                chunk_start = i + 1
        if rows and rows[-1]:
            self.importChunk(rows[chunk_start:], chunk_start + 3, dry_run)

    def listTerms(self):
        first, last = self.terms
        result = [first]
        next = getNextTerm(first)
        while (next is not None and
               not sameProxiedObjects(result[-1], next)):
            result.append(next)
            next = getNextTerm(next)
        return result

    def importChunk(self, rows, line, dry_run=True):
        """Import a chunk of data that describes a section.

        You should run this method with dry_run=True before trying the
        real thing, or you might get in trouble.
        """
        terms = self.listTerms()
        row = rows[0]
        if len(row) not in (2, 2 + len(terms)):
            err_msg = _('Wrong section header on line ${line_no} (it should contain a'
                        ' course id, an instructor id and optional SchoolTool '
                        'section IDs for each of the terms)',
                        mapping={'line_no': line})
            return

        section_ids = None
        if len(row) == 2:
            course_id, instructor_id = row[:2]
        else:
            course_id = row[0]
            instructor_id = row[1]
            section_ids = row[2:]

        course = ICourseContainer(self.schoolyear).get(course_id, None)
        if course is None:
            self.errors.courses.append(course_id)

        instructor = self.persons.get(instructor_id, None)
        if instructor is None:
            self.errors.persons.append(instructor_id)

        line_ofs = 1
        finished = False

        timetables = ITimetableContainer(self.schoolyear)
        timetable = None
        periods = {}

        for row in rows[1:]:
            line_ofs += 1
            if row == ['***']:
                finished = True
                break

            if len(row) == 1:
                tt = timetables.get(row[0])
                if tt is None:
                    err_msg = _("Malformed line ${line_no}"
                                " (it should contain either a timetable id or"
                                " day id and a period id)",
                                mapping={'line_no': line + line_ofs - 1})
                    self.errors.generic.append(err_msg)
                    continue
                timetable = tt
                continue
            elif len(row) == 2:
                day_id, period_id = row
            else:
                err_msg = _("Malformed line ${line_no}"
                            " (it should contain either a timetable id or"
                            " day id and a period id)",
                            mapping={'line_no': line + line_ofs - 1})
                self.errors.generic.append(err_msg)
                continue

            if timetable is None:
                err_msg = _("Timetable id must be specified before"
                            " day id and a period id"
                            " at at line ${line_no}",
                            mapping={'line_no': line + line_ofs - 1})
                continue

            # check day_id
            ttday = None
            for day in timetable.periods.templates.values():
                if day.title == day_id:
                    ttday = day
                    break

            if ttday is None:
                errkey = (timetable.__name__, day_id)
                if errkey not in self.errors.day_ids:
                    self.errors.day_ids.append(errkey)
                continue

            ttperiod = None
            for period in ttday.values():
                if period.title == period_id:
                    ttperiod = period
                    break

            # check period_id
            if ttperiod is None:
                errkey = (timetable.__name__, day_id, period_id)
                if period_id not in self.errors.periods:
                    self.errors.periods.append(errkey)
                    continue

            if timetable.__name__ not in periods:
                periods[timetable.__name__] = []
            periods[timetable.__name__].append(period)

        if not finished:
            err_msg = _("Incomplete section description on line ${line}",
                        mapping = {'line': line})
            self.errors.generic.append(err_msg)
            return
        if len(rows) == line_ofs:
            err_msg = _("No students in section (line ${line})",
                        mapping = {'line': line + line_ofs})
            self.errors.generic.append(err_msg)
            return


        sections = []
        for n, term in enumerate(terms):
            section_container = ISectionContainer(term)
            section_id = None
            if section_ids is not None:
                section_id = section_ids[n]
            if (section_id is not None and
                section_id in section_container):
                section = section_container[section_id]
                self.updateSection(
                    section, term, course, instructor, periods,
                    dry_run=dry_run)
            else:
                section = self.createSection(
                    term, course, instructor, periods,
                    section_id=section_id,
                    dry_run=dry_run)
            self.importPersons(rows[line_ofs:], section, dry_run=dry_run)
            if section is not None:
                sections.append(section)
        if not self.errors.anyErrors():
            for n, section in enumerate(sections[:-1]):
                section.next = sections[n+1]

    def createSection(self, term, course, instructor, periods,
                      section_id=None, dry_run=True):
        """Create a section.

        `periods` is a list of tuples (day_id, period_id).

        A title is generated from the titles of `course` and `instructor`.
        If an existing section with the same title is found, it is used instead
        of creating a new one.

        The created section is returned, or None if dry_run is True.
        """
        if dry_run:
            return None

        sections = ISectionContainer(term)

        section = Section()
        chooser = INameChooser(sections)
        auto_name = chooser.chooseName('', section)
        section.title = u"%s (%s)" % (course.title, auto_name)
        if section_id is None:
            section_id = auto_name
        sections[section_id] = section

        # Establish links to course and to teacher
        if course not in section.courses:
            section.courses.add(course)
        if instructor not in section.instructors:
            section.instructors.add(instructor)

        timetable_container = ITimetableContainer(self.schoolyear)
        timetables = [timetable_container[ttid]
                      for ttid in sorted(periods)]
        schedules = IScheduleContainer(section)
        for timetable in timetables:
            selected = periods[timetable.__name__]
            schedule = SelectedPeriodsSchedule(
                timetable, term.first, term.last,
                title=timetable.title, timezone=timetable.timezone)
            for period in selected:
                schedule.addPeriod(period)
            schedules[timetable.__name__] = schedule

        return section

    def updateSection(self, section, term, course, instructor, periods,
                      dry_run=True):
        """Create a section.

        `periods` is a list of tuples (day_id, period_id).

        A title is generated from the titles of `course` and `instructor`.
        If an existing section with the same title is found, it is used instead
        of creating a new one.

        The created section is returned, or None if dry_run is True.
        """
        if dry_run:
            return None

        # Establish links to course and to teacher
        for c in list(section.courses):
            if c is not course:
                section.remove(c)
        if course not in section.courses:
            section.courses.add(course)
        for i in list(section.instructors):
            if i is not instructor:
                section.instructor.remove(i)
        if instructor not in section.instructors:
            section.instructors.add(instructor)

        timetable_container = ITimetableContainer(self.schoolyear)
        timetables = [timetable_container[ttid]
                      for ttid in sorted(periods)]
        schedules = IScheduleContainer(section)
        for timetable in timetables:
            selected = periods[timetable.__name__]
            schedule = None
            for s in schedules.values():
                if sameProxiedObjects(s.timetable, timetable):
                    schedule = s
                    break
            if schedule is None:
                schedule = SelectedPeriodsSchedule(
                    timetable, term.first, term.last,
                    title=timetable.title, timezone=timetable.timezone)
                for period in selected:
                    schedule.addPeriod(period)
                schedules[timetable.__name__] = schedule
            else:
                for period in schedule.periods:
                    if period not in selected:
                        schedule.removePeriod(period)
                for period in selected:
                    schedule.addPeriod(period)

    def importPersons(self, person_data, section, dry_run=True):
        """Import persons into a section."""
        for row in person_data:
            person_id = row[0]
            try:
                person = self.persons[person_id]
            except KeyError:
                if person_id not in self.errors.persons:
                    self.errors.persons.append(person_id)
            else:
                if not dry_run:
                    if person not in section.members:
                        section.members.add(person)

    def findTerm(self, title):
        terms = ITermContainer(self.schoolyear)
        for term in terms.values():
            if term.title == title:
                return term
        return None

    def importHeader(self, row):
        """Read the header row of the CSV file.

        Sets self.terms
        """
        if len(row) < 1 or len(row) > 2:
            self.errors.generic.append(
                    _("The first row of the CSV file must contain"
                      " the term or first and last terms for sections."))
            return

        first = self.findTerm(row[0])

        if first is None:
            error_msg = _("The term ${term} does not exist.",
                          mapping={'term': row[0]})
            self.errors.generic.append(error_msg)
            return

        if len(row) == 1:
            last = first
        else:
            last = self.findTerm(row[1])
            if last is None:
                error_msg = _("The term ${term} does not exist.",
                              mapping={'term': row[1]})
                self.errors.generic.append(error_msg)
                return

        if first.first >= last.last:
                error_msg = _("First term ${first} starts after"
                              " last term ${last}.",
                              mapping={'first': row[0],
                                       'last': row[1]})
                self.errors.generic.append(error_msg)
                return

        self.terms = (first, last)


    def parseCSVRows(self, rows):
        """Parse rows (a list of strings) in CSV format.

        Returns a list of rows as lists.  Trailing empty cells are discarded.

        rows must be in the encoding specified during construction of
        TimetableCSVImportView; the returned values are in unicode.

        If the provided data is invalid, self.errors.generic will be updated
        and None will be returned.
        """
        result = []
        reader = csv.reader(rows)
        line = 0
        try:
            while True:
                line += 1
                values = [v.strip() for v in reader.next()]
                if self.charset:
                    values = [unicode(v, self.charset) for v in values]
                # Remove trailing empty cells.
                while values and not values[-1].strip():
                    del values[-1]
                result.append(values)
        except StopIteration:
            return result
        except csv.Error:
            error_msg = _("Error in timetable CSV data, line ${line_no}",
                          mapping={'line_no': line})
            self.errors.generic.append(error_msg)
            return
        except UnicodeError:
            error_msg = _("Conversion to unicode failed in line ${line_no}",
                          mapping={'line_no': line})
            self.errors.generic.append(error_msg)
            return


class TimetableCSVImportView(BaseCSVImportView):
    """Timetable CSV import view."""

    __used_for__ = ISectionContainer

    importer_class = TimetableCSVImporter

    def _presentErrors(self, err):
        if err.generic:
            self.errors.extend(err.generic)

        for key, msg in [
            ('day_ids', _("Day ids not defined in selected schema: ${args}.")),
            ('periods', _("Periods not defined in selected days: ${args}.")),
            ('persons', _("Persons not found: ${args}.")),
            ('courses', _("Courses not found: ${args}.")),
            ('sections', _("Sections not found: ${args}.")),
            ('records', _("Invalid records: ${args}."))]:
            v = getattr(err, key)
            if v:
                values = ', '.join([unicode(st) for st in v])
                msg = _(msg, mapping={'args': values})
                self.errors.append(msg)


class FlourishTimetableCSVImportView(FlourishBaseCSVImportView,
                                     TimetableCSVImportView,
                                     ActiveSchoolYearContentMixin):

    @property
    def schoolyear(self):
        return ISchoolYear(self.context)

    def nextURL(self):
        app = ISchoolToolApplication(None)
        return self.url_with_schoolyear_id(app, view_name='sections')

    @property
    def title(self):
        return _('Sections for ${schoolyear}',
                 mapping={'schoolyear': self.schoolyear.title})


class ImportSectionsLinkViewlet(flourish.page.LinkViewlet,
                                ActiveSchoolYearContentMixin):

    @property
    def url(self):
        link = self.link
        if not link:
            return None
        return "%s/%s" % (absoluteURL(self.schoolyear, self.request),
                          self.link)



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2011 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from zope.app.generations.utility import findObjectsProviding
from zope.annotation.interfaces import IAnnotatable, IAnnotations
from zope.component import getUtility
from zope.intid.interfaces import IIntIds

from schooltool.app.overlay import URICalendarSubscriber
from schooltool.app.overlay import CalendarOverlayInfo
from schooltool.course.interfaces import ISection
from schooltool.generations.evolve36.helper import assert_not_broken
from schooltool.generations.evolve36.helper import BuildContext
from schooltool.generations.evolve36.model import ITimetableCalendarEvent
from schooltool.relationship import relate, unrelate
from schooltool.relationship.interfaces import IRelationshipLinks
from schooltool.timetable.calendar import ScheduleCalendar

ST_CALENDAR_KEY = 'schooltool.app.calendar.Calendar'
SCHEDULE_CALENDAR_KEY = 'schooltool.timetable.app.ScheduleCalendar'


def getScheduleCalendar(owner):
    if not ISection.providedBy(owner):
        return None
    annotations = IAnnotations(owner)
    try:
        return annotations[SCHEDULE_CALENDAR_KEY]
    except KeyError:
        calendar = ScheduleCalendar(owner)
        annotations[SCHEDULE_CALENDAR_KEY] = calendar
        return calendar


class CalendarBuilder(object):

    events = None
    calendar = None

    store_data = ('__name__', 'first', 'last',
                   'timezone', 'term',
                   'consecutive_periods_as_one')

    def read(self, calendar, context):
        self.calendar = calendar
        self.events = []
        for event in calendar:
            assert_not_broken(event)
            if not ITimetableCalendarEvent.providedBy(event):
                continue
            if event.__parent__ != calendar:
                # Event does not belong directly to this calendar
                # Probably this is a booked resource
                continue
            activity = event.activity

            schema = activity.timetable.schooltt
            year_int_id = int(schema.__parent__.__name__)
            period_key = (year_int_id,
                          schema.__name__,
                          event.day_id,
                          event.period_id)

            timetable_key = (activity.timetable.term,
                             activity.owner,
                             activity.timetable.__name__)

            self.events.append({
                    'dtstart': event.dtstart,
                    'duration': event.duration,
                    'unique_id': event.unique_id,
                    'timetable_key': timetable_key,
                    'period_key': period_key,
                    # copy
                    'description': event.description,
                    'location': event.location,
                    'resources': event.resources,
                    })

    def clean(self, context):
        for event_info in self.events:
            event = self.calendar.find(event_info['unique_id'])
            self.calendar.removeEvent(event)

    def findEvent(self, calendar, period, dtstart, duration):
        for event in calendar:
            if (event.period is period and
                event.dtstart == dtstart and
                event.duration == duration):
                return event
        return None

    def build(self, context):
        int_ids = getUtility(IIntIds)
        schedule_calendars = set()
        calendar = getScheduleCalendar(self.calendar.__parent__)
        if calendar is None:
            return

        schedule_calendars.add(calendar)

        for event in self.events:
            schedule = context.shared.schedule_map.get(event['timetable_key'])
            if schedule is None:
                continue

            period = context.shared.period_map[event['period_key']]
            owner_int_id = int(schedule.__parent__.__name__)
            owner = int_ids.getObject(owner_int_id)

            calendar = getScheduleCalendar(owner)
            if calendar is None:
                continue

            if calendar not in schedule_calendars:
                schedule_calendars.add(calendar)

            new_event = self.findEvent(
                calendar, period, event['dtstart'], event['duration'])
            if new_event is not None:
                new_event.description = event['description']
                new_event.location = event['location']
                for resource in event['resources']:
                    if resource not in new_event.resources:
                        new_event.bookResource(resource)

        # XXX: maybe copy over old "free" section events here

        schedule_cal_relationships = [
            (cal, IRelationshipLinks(cal)) for cal in schedule_calendars]

        old_relationships = IRelationshipLinks(self.calendar)
        old_subscriptions = list(
            old_relationships.getLinksByRole(URICalendarSubscriber))

        for link in old_subscriptions:
            old_info = link.extra_info
            for schedule_cal, relationships in schedule_cal_relationships:
                info = CalendarOverlayInfo(schedule_cal, old_info.show,
                                           old_info.color1, old_info.color2)
                info.__parent__ = old_info.__parent__
                try:
                    relationships.find(
                        link.my_role, link.target, link.role, link.rel_type)
                except ValueError:
                    relate(link.rel_type,
                           (schedule_cal, link.my_role),
                           (link.target, link.role),
                           extra_info = info)

        for link in old_subscriptions:
            unrelate(link.rel_type,
                     (self.calendar, link.my_role),
                     (link.target, link.role))


class AppTimetableCalendarBuilder(object):
    builders = None

    def read(self, app, context):
        self.builders = []
        candidates = findObjectsProviding(app, IAnnotatable)
        for candidate in candidates:
            assert_not_broken(candidate)
            annotations = IAnnotations(candidate, None)
            if annotations is None:
                continue
            calendar = annotations.get(ST_CALENDAR_KEY)
            if calendar is None:
                continue
            assert_not_broken(calendar)
            builder = CalendarBuilder()
            builder.read(calendar, context(app=app))
            self.builders.append(builder)

    def clean(self, app, context):
        for builder in self.builders:
            builder.clean(context)

    def build(self, app, context):
        result = BuildContext()
        for builder in self.builders:
            built = builder.build(context(app=app))
        return result



#
# SchoolTool - common information systems platform for school administration
# Copyright (c) 2008 Shuttleworth Foundation
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Adapter subscriber
"""
from zope.interface import Interface
from zope.interface import implements
from zope.component.interfaces import IObjectEvent
from zope.component import adapter
from zope.component import adapts
from zope.component import getAdapters

from schooltool.schoolyear.interfaces import ISubscriber

@adapter(Interface)
def subscriberAdapterDispatcher(event):
    subscribers = getAdapters((event,), ISubscriber)
    for name, subscriber in subscribers:
        subscriber()


class EventAdapterSubscriber(object):
    implements(ISubscriber)

    def __init__(self, event):
        self.event = event

    def __call__(self):
        raise NotImplementedError("Please override this method in subclasses")


class ObjectEventAdapterSubscriberDispatcher(EventAdapterSubscriber):
    adapts(IObjectEvent)
    implements(ISubscriber)

    def __call__(self):
        subscribers = getAdapters((self.event, self.event.object), ISubscriber)
        for name, subscriber in subscribers:
            subscriber()


class ObjectEventAdapterSubscriber(object):
    implements(ISubscriber)

    def __init__(self, event, object):
        self.event = event
        self.object = object

    def __call__(self):
        raise NotImplementedError("Please override this method in subclasses")
